"""Flformdb module."""

# -*- coding: utf-8 -*-
import traceback
from PyQt6 import QtCore, QtGui, QtWidgets  # type: ignore[import]

from pineboolib import logging

from pineboolib.core import decorators, settings
from pineboolib.core.utils import utils_base

from pineboolib.application.utils import geometry
from pineboolib.application.metadata import pnaction

from pineboolib.q3widgets import qmainwindow

from pineboolib.application.database import pnsqlcursor

from pineboolib import application
from typing import Any, Union, Dict, Optional, Tuple, Type, cast, Callable, TYPE_CHECKING


if TYPE_CHECKING:
    from pineboolib.interfaces import isqlcursor  # pragma: no cover
    from pineboolib.qsa import formdbwidget  # pragma: no cover


LOGGER = logging.get_logger(__name__)


class FLFormDB(QtWidgets.QDialog):
    """
    Represents a form that links to a table.

    It is used as a container of components that want
    link to the database and access the records
    of the cursor. This structure greatly simplifies
    measure access to data since many tasks are
    Automatically managed by this container form.

    At first the form is created empty and we must invoke
    the FLFormDB :: setMainWidget () method, passing it as a parameter
    another widget (usually a form created with QtDesigner),
    which contains different components, this widget will be displayed
    inside this container, self-configuring all the components
    It contains, with the cursor data and metadata. Generally the
    Components will be plugins, such as FLFieldDB or FLTableDB
    """

    """
    Cursor, con los registros, utilizado por el formulario
    """
    cursor_: Optional["isqlcursor.ISqlCursor"]

    """
    Nombre de la tabla, contiene un valor no vacío cuando
    la clase es propietaria del cursor
    """
    name_: str

    """
    Capa principal del formulario
    """
    layout_: "QtWidgets.QVBoxLayout"

    """
    Widget principal del formulario
    """
    main_widget: Optional["QtWidgets.QWidget"]
    """
    Identificador de ventana MDI.

    Generalmente es el nombre de la acción que abre el formulario
    """
    _id_mdi: str

    """
    Capa para botones
    """
    layoutButtons: "QtWidgets.QHBoxLayout"

    """
    Boton Cancelar
    """
    pushButtonCancel: Optional["QtWidgets.QToolButton"]

    """
    Indica que la ventana ya ha sido mostrada una vez
    """
    _showed: bool

    """
    Guarda el contexto anterior que tenia el cursor
    """
    _old_cursor_context: Any

    """
    Indica que el formulario se está cerrando
    """
    _is_closing: bool

    """
    Componente con el foco inicial
    """
    _init_focus_widget: Optional["QtWidgets.QWidget"]

    """
    Guarda el último objeto de formulario unido a la interfaz de script (con bindIface())
    """
    _old_form_object: Any

    """
    Boton Debug Script
    """
    pushButtonDebug: Optional["QtWidgets.QToolButton"]

    """
    Almacena que se aceptado, es decir NO se ha pulsado, botón cancelar
    """
    accepted_: bool

    """
    Nombre del formulario relativo a la acción (form / formRecrd + nombre de la acción)
    """
    _action_name: str

    """
    Interface para scripts
    """
    # iface: Any

    """
    Tamaño de icono por defecto
    """
    _icon_size: "QtCore.QSize"

    # protected slots:

    """
    Uso interno
    """
    _old_form_objectDestroyed = QtCore.pyqtSignal()

    # signals:

    """
    Señal emitida cuando se cierra el formulario
    """
    closed = QtCore.pyqtSignal()

    """
    Señal emitida cuando el formulario ya ha sido inicializado y está listo para usarse
    """
    formReady = QtCore.pyqtSignal()
    formClosed = QtCore.pyqtSignal()

    known_instances: Dict[Tuple[Type["FLFormDB"], str], "FLFormDB"] = {}

    bottomToolbar: QtWidgets.QFrame

    toolButtonClose: Optional["QtWidgets.QToolButton"]

    _ui_name: str

    loop: bool
    _action: "pnaction.PNAction"

    eventloop: "QtCore.QEventLoop"

    def __init__(
        self,
        action_or_name: Union["pnaction.PNAction", str],
        parent: Optional[Union["QtWidgets.QWidget", int]] = None,
        load: Union[bool, int] = False,
    ) -> None:
        """Create a new FLFormDB for given action."""
        # self.tiempo_ini = time.time()
        parent_widget: QtWidgets.QWidget

        if isinstance(load, int):
            load = load == 1

        if parent is None or isinstance(parent, int):
            if application.PROJECT.main_window is not None:
                if application.PROJECT.main_window.main_widget is not None:
                    parent_widget = application.PROJECT.main_window.main_widget
                else:
                    parent_widget = application.PROJECT.main_window
            else:
                raise Exception("main_window is not loaded!")
        else:
            parent_widget = parent

        super().__init__(parent_widget)

        self._loaded = False

        if isinstance(action_or_name, str):
            self._action = application.PROJECT.conn_manager.manager().action(action_or_name)
        else:
            self._action = action_or_name

        self.known_instances[(self.__class__, self._action.name())] = self

        self._ui_name = self._action_name = self._action.name()

        if self._action.table():
            if type(self).__name__ == "FLFormRecordDB":
                self._action_name = "formRecord%s" % self._action_name
                # script_name = self._action.scriptFormRecord()
                # self.action_widget = application.PROJECT.actions[self._action.name()]._record_widget
            else:
                self._action_name = "form%s" % self._action_name
                # script_name = self._action.scriptForm()
                # self.action_widget = application.PROJECT.actions[self._action.name()]._master_widget
            self._ui_name = self._action.form()
        else:
            if self._action.scriptForm() or self._action.scriptFormRecord():
                self._ui_name = ""

        # self.mod = self._action.mod
        self.loop = False
        self.eventloop = QtCore.QEventLoop()

        self.layout_ = QtWidgets.QVBoxLayout()
        self.layout_.setContentsMargins(1, 1, 1, 1)
        self.layout_.setSpacing(1)
        self.layout_.setContentsMargins(1, 1, 1, 1)
        self.layout_.setSizeConstraint(QtWidgets.QLayout.SizeConstraint.SetMinAndMaxSize)
        self.setLayout(self.layout_)

        self.pushButtonCancel = None  # pylint: disable=invalid-name
        self.toolButtonClose = None  # pylint: disable=invalid-name
        self.bottomToolbar = QtWidgets.QFrame()  # pylint: disable=invalid-name
        # self.cursor_ = None
        self._init_focus_widget = None
        self._showed = False
        self._is_closing = False
        self.accepted_ = False
        self.main_widget = None
        # self.iface = None
        self._old_form_object = None
        self._old_cursor_context = None

        self._id_mdi = self._action.name()
        self._icon_size = application.PROJECT.DGI.icon_size()

        if load:
            self.load()
            self.initForm()

    def load(self) -> None:
        """Load control."""
        if self._loaded:
            return

        # self.resize(550,350)
        if self.layout_ is None:
            return

        widget: Union["qmainwindow.QMainWindow", "QtWidgets.QDialog"]

        if not self._action.table():
            widget = qmainwindow.QMainWindow()
        else:
            widget = QtWidgets.QDialog()

        # LOGGER.warning("previo %s %s %s", self._action.form(), self._action.formRecord(), widget)

        self.layout_.insertWidget(0, widget)
        self.layout_.setSpacing(1)
        self.layout_.setContentsMargins(1, 1, 1, 1)
        self.layout_.setSizeConstraint(QtWidgets.QLayout.SizeConstraint.SetMinAndMaxSize)
        if self._ui_name:
            if application.PROJECT.conn_manager is None:
                raise Exception("Project is not connected yet")

            application.PROJECT.conn_manager.managerModules().createUI(self._ui_name, None, widget)

        self._loaded = True

    def loaded(self) -> bool:
        """Return if the control is initialized."""

        return self._loaded

    @decorators.pyqt_slot()
    def initScript(self) -> bool:
        """
        Call the "init" function of the masterprocess script associated with the form.
        """
        if self._loaded:
            if self.action_widget:
                self.action_widget.clear_connections()

            fun = getattr(self.iface, "init", None)
            if fun is None:
                fun = getattr(self.action_widget, "init", None)

            if fun is not None:
                fun()

            return True

        return False

    # def __del__(self) -> None:
    #    """
    #    Destroyer.
    #    """
    #    # TODO: Esto hay que moverlo al closeEvent o al close()
    #    # ..... los métodos __del__ de python son muy poco fiables.
    #    # ..... Se lanzan o muy tarde, o nunca.
    #    # (De todos modos creo que ya hice lo mismo a mano en el closeEvent en commits anteriores)
    #
    #    self.unbindIface()

    def setCursor(self, cursor: Optional["isqlcursor.ISqlCursor"] = None) -> None:  # type: ignore
        """Change current cursor binded to this control."""
        if cursor is None:
            return

        if cursor is not self.cursor_:
            if self.cursor_ is not None:
                if self._old_cursor_context:
                    self.cursor_.setContext(self._old_cursor_context)

                if type(self).__name__ == "FLFormRecodDB":
                    self.cursor_.restoreEditionFlag(self.objectName())
                    self.cursor_.restoreBrowseFlag(self.objectName())

            # if self.cursor_:

            #    cast(QtCore.pyqtSignal, self.cursor_.destroyed).disconnect(self.cursorDestroyed)

            # self.widget.cursor_ = cursor
            self.cursor_ = cursor

            if type(self).__name__ == "FLFormRecodDB":
                self.cursor_.setEdition(False, self.objectName())
                self.cursor_.setBrowse(False, self.objectName())

            # cast(QtCore.pyqtSignal, self.cursor_.destroyed).connect(self.cursorDestroyed)
            iface = self.iface

            if iface is not None and self.cursor_ is not None:
                self._old_cursor_context = self.cursor_.context()
                self.cursor_.setContext(self.iface)

    def cursor(self) -> "isqlcursor.ISqlCursor":  # type: ignore [override] # noqa F821
        """
        To get the cursor used by the form.
        """
        if self.cursor_ is None:
            raise Exception("cursor_ is empty!.")

        return self.cursor_

    def mainWidget(self) -> Optional["QtWidgets.QWidget"]:
        """
        To get the form's main widget.
        """

        return self.main_widget

    def setIdMDI(self, id_: str) -> None:
        """
        Set the MDI ID.
        """

        self._id_mdi = id_

    def idMDI(self) -> str:
        """
        Return the MDI ID.
        """

        return self._id_mdi

    def setMainWidget(self, widget: Optional["QtWidgets.QWidget"] = None) -> None:
        """
        Set widget as the main form.
        """
        if widget is not None:
            self.main_widget = widget
        else:
            self.main_widget = self

    def snapShot(self) -> "QtGui.QImage":
        """
        Return the image or screenshot of the form.
        """
        pix = self.grab()
        return pix.toImage()

    def saveSnapShot(self, path_file: Optional[str] = None) -> None:
        """
        Save the image or screenshot of the form in a PNG format file.
        """
        if not path_file:
            tmp_file = "%s/snap_shot_%s.png" % (
                application.PROJECT.tmpdir,
                QtCore.QDateTime.currentDateTime().toString("ddMMyyyyhhmmsszzz"),
            )

            ret = QtWidgets.QFileDialog.getSaveFileName(
                QtWidgets.QApplication.activeWindow(), "Pineboo", tmp_file, "PNG(*.png)"
            )
            path_file = ret[0] if ret else None

        if path_file:
            file_ = QtCore.QFile(path_file)
            if file_.openMode() != QtCore.QIODevice.OpenModeFlag.WriteOnly:
                self.tr("Error I/O al intentar escribir el fichero %s" % path_file)
                return

            self.snapShot().save(file_, "PNG")

    def saveGeometry(self) -> "QtCore.QByteArray":
        """Save current window size into settings."""
        # pW = self.parentWidget()
        # if not pW:
        geo = QtCore.QSize(self.width(), self.height())
        if self.isMinimized():
            geo.setWidth(1)
        elif self.isMaximized():
            geo.setWidth(9999)
        # else:
        #    geo = QtCore.QSize(pW.width(), pW.height())

        geometry.save_geometry_form(self.geoName(), geo)
        return super().saveGeometry()

    def setCaptionWidget(self, text: str) -> None:
        """
        Set the window title.
        """
        if not text:
            return

        self.setWindowTitle(text)

    def accepted(self) -> bool:  # type: ignore
        """
        Return if the form has been accepted.
        """
        # FIXME: QtWidgets.QDialog.accepted() is a signal. We're shadowing it.
        return self.accepted_

    def formClassName(self) -> str:
        """
        Return the class name of the form at runtime.
        """
        return "FormDB"

    def exec_(self) -> bool:
        """
        Only to be compatible with FLFormSearchDB. By default, just call QWidget.show.
        """

        super().show()
        return True

    def hide(self) -> None:
        """Hide control."""
        super().hide()

    @decorators.pyqt_slot()
    def close(self) -> bool:
        """
        Close the form.
        """
        if self._is_closing or not self._loaded:
            return True

        self._is_closing = True

        super().close()
        self._is_closing = False
        return True

    @decorators.pyqt_slot()
    def accept(self) -> None:
        """
        Activated by pressing the accept button.
        """
        pass

    @decorators.pyqt_slot()
    def reject(self) -> None:
        """
        Activated by pressing the cancel button.
        """
        pass

    @decorators.pyqt_slot()
    def showForDocument(self) -> None:
        """
        Show the form without calling the script "init".

        Used in documentation to avoid conflicts when capturing forms.
        """
        self._showed = True
        if self.main_widget:
            self.main_widget.show()
            self.resize(self.main_widget.size())
        super().show()

    @decorators.pyqt_slot()
    @decorators.not_implemented_warn
    def debugScript(self) -> bool:
        """
        Show the script associated with the form in the Workbench to debug.
        """

        return True

    @decorators.pyqt_slot()
    def get_script(self) -> Optional[str]:
        """
        Return the script associated with the form.
        """

        ifc = self.iface
        if ifc:
            return str(ifc)
        return None

    # private slots:

    @decorators.pyqt_slot()
    def callInitScript(self) -> None:
        """Call QS Script related to this control."""
        if not self.initScript():
            raise Exception("Error initializing the module.")

        if not self._is_closing:
            QtCore.QTimer.singleShot(0, self.emitFormReady)

    def emitFormReady(self) -> None:
        """Emit formReady signal, after the form has been loaded."""

        if "fltesttest" in application.PROJECT.conn_manager.managerModules().listAllIdModules():
            application.PROJECT.call(
                "fltesttest.iface.recibeEvento", ["formReady", self._action_name], None
            )
        self.formReady.emit()

    # protected_:

    def emitFormClosed(self) -> None:
        """Emit formClosed signal."""

        if application.PROJECT.conn_manager is None:
            raise Exception("Project is not connected yet")

        if "fltesttest" in application.PROJECT.conn_manager.managerModules().listAllIdModules():
            application.PROJECT.call(
                "fltesttest.iface.recibeEvento", ["formClosed", self._action_name], None
            )

        self.formClosed.emit()

        widget = self.action_widget

        if widget is not None:
            self.action_widget.closed.emit()

    def action(self) -> "pnaction.PNAction":
        """Get form PNAction."""
        return self._action

    def initForm(self) -> None:
        """
        Initialize the associated script.
        """

        acl = application.PROJECT.aq_app.acl()

        if acl:
            acl.process(self)

        self.loadControls()

        if self._action is None:
            raise Exception("_action is empty!")

        if self._action.table():
            if (
                not self.cursor_
                or not self.cursor_._action
                or self.cursor_._action.table() is not self._action.table()
            ):
                cursor = pnsqlcursor.PNSqlCursor(self._action.table())
                self.setCursor(cursor)

            # if self._loaded and not self.__class__.__name__ == "FLFormRecordDB":
            # application.PROJECT.conn_manager.managerModules().loadFLTableDBs(self)

            if self._action.description() not in ("", None):
                self.setWhatsThis(self._action.description())

            caption = self._action.caption()

            if caption in ("", None) and self.cursor_ and self.cursor_.metadata():
                caption = self.cursor_.metadata().alias()

            if caption in ("", None):
                caption = QtWidgets.QApplication.translate("FLFormDB", "No hay metadatos")
            self.setCaptionWidget(caption)

    def loadControls(self) -> None:
        """Load form controls."""

        if self.pushButtonCancel:
            self.pushButtonCancel.hide()

        if self.bottomToolbar and self.toolButtonClose:
            self.toolButtonClose.hide()

        self.init_tool_bar()

        size_policy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy(0), QtWidgets.QSizePolicy.Policy(0)
        )
        size_policy.setHeightForWidth(True)

        push_button_size = self._icon_size

        if settings.CONFIG.value("application/isDebuggerMode", False):
            pushButtonExport = QtWidgets.QToolButton()  # pylint: disable=invalid-name
            pushButtonExport.setObjectName("pushButtonExport")
            pushButtonExport.setSizePolicy(size_policy)
            pushButtonExport.setMinimumSize(push_button_size)
            pushButtonExport.setMaximumSize(push_button_size)
            pushButtonExport.setIcon(
                QtGui.QIcon(utils_base.filedir("./core/images/icons", "gtk-properties.png"))
            )
            pushButtonExport.setShortcut(QtGui.QKeySequence(self.tr("F3")))
            pushButtonExport.setWhatsThis(
                QtWidgets.QApplication.translate("FLFormDB", "Exportar a XML(F3)")
            )
            pushButtonExport.setToolTip(
                QtWidgets.QApplication.translate("FLFormDB", "Exportar a XML(F3)")
            )
            pushButtonExport.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
            self.bottomToolbar.layout().addWidget(pushButtonExport)  # type: ignore [union-attr]
            pushButtonExport.clicked.connect(  # type: ignore [attr-defined] # noqa: F821
                self.exportToXml
            )

            if settings.CONFIG.value("ebcomportamiento/show_snaptshop_button", False):
                push_button_snapshot = QtWidgets.QToolButton()
                push_button_snapshot.setObjectName("pushButtonSnapshot")
                push_button_snapshot.setSizePolicy(size_policy)
                push_button_snapshot.setMinimumSize(push_button_size)
                push_button_snapshot.setMaximumSize(push_button_size)
                push_button_snapshot.setIcon(
                    QtGui.QIcon(utils_base.filedir("./core/images/icons", "gtk-paste.png"))
                )
                push_button_snapshot.setShortcut(QtGui.QKeySequence(self.tr("F8")))
                push_button_snapshot.setWhatsThis("Capturar pantalla(F8)")
                push_button_snapshot.setToolTip("Capturar pantalla(F8)")
                push_button_snapshot.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
                self.bottomToolbar.layout().addWidget(push_button_snapshot)  # type: ignore [union-attr]
                push_button_snapshot.clicked.connect(  # type: ignore [attr-defined] # noqa: F821
                    self.saveSnapShot
                )

            spacer = QtWidgets.QSpacerItem(
                20, 20, QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed
            )
            self.bottomToolbar.layout().addItem(spacer)  # type: ignore [union-attr]

        if not self.pushButtonCancel:
            self.pushButtonCancel = QtWidgets.QToolButton()
            self.pushButtonCancel.setObjectName("pushButtonCancel")
            cast(
                QtCore.pyqtSignal, self.pushButtonCancel.clicked
            ).connect(  # type: ignore [attr-defined] # noqa: F821
                cast(Callable, self.close)
            )

        self.pushButtonCancel.setSizePolicy(size_policy)
        self.pushButtonCancel.setMaximumSize(push_button_size)
        self.pushButtonCancel.setMinimumSize(push_button_size)
        self.pushButtonCancel.setIcon(
            QtGui.QIcon(utils_base.filedir("./core/images/icons", "gtk-stop.png"))
        )
        # self.pushButtonCancel.setFocusPolicy(QtCore.Qt.StrongFocus)
        # self.pushButtonCancel.setFocus()
        self.pushButtonCancel.setShortcut(QtGui.QKeySequence(self.tr("Esc")))
        self.pushButtonCancel.setWhatsThis("Cerrar formulario (Esc)")
        self.pushButtonCancel.setToolTip("Cerrar formulario (Esc)")
        self.bottomToolbar.layout().addWidget(self.pushButtonCancel)  # type: ignore [union-attr]
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)

    def formName(self) -> str:
        """
        Return internal form name.
        """

        return "form%s" % self._id_mdi

    def name(self) -> str:
        """Get name of the form."""

        return self.formName()

    def geoName(self) -> str:
        """Get name of the form."""
        # FIXME: What this should do exactly?
        return self.formName()

    # def bindIface(self) -> None:
    #    """
    #    Join the script interface to the form object.
    #    """

    #    if self.iface:
    #        self._old_form_object = self.iface

    # def unbindIface(self) -> None:
    #    """
    #    Disconnect the script interface to the form object.
    #    """
    #    if not self.iface:
    #        return

    #    self.iface = self._old_form_object

    # def isIfaceBind(self) -> bool:
    #    """
    #    Indicate if the script interface is attached to the form object.
    #    """

    #    if self.iface:
    #        return True
    #    else:
    #        return False

    def closeEvent(self, event: Optional["QtGui.QCloseEvent"]) -> None:
        """
        Capture event close.
        """

        self.frameGeometry()

        self.saveGeometry()

        self.hide()
        self.emitFormClosed()
        self._loaded = False

        # from PyQt6.QtWidgets import qApp

        # qApp.processEvents() #Si se habilita pierde mucho tiempo!

        # self.hide()
        try:
            if type(self).__name__ != "FLFormSearchDB":
                super().close()

            instance_name = (self.__class__, self._action.name())
            if instance_name in self.known_instances.keys():
                del self.known_instances[instance_name]

        except Exception:
            LOGGER.error(
                "El FLFormDB %s no se cerró correctamente:\n%s",
                self.formName(),
                traceback.format_exc(),
            )

        parent = self.parent()

        if isinstance(parent, QtWidgets.QMdiSubWindow):
            parent.close()

    def showEvent(self, event: Optional["QtGui.QShowEvent"] = None) -> None:
        """
        Capture event show.
        """
        # --> Para mostrar form sin negro previo
        # QtWidgets.QApplication.processEvents()
        # <--

        if not self.loaded():
            return

        if not self._showed:
            self._showed = True

            size = geometry.load_geometry_form(self.geoName())
            if size:
                self.resize(size)

            parent = self.parent()

            if parent and isinstance(parent, QtWidgets.QMdiSubWindow):
                if size:
                    parent.resize(size)
                parent.repaint()

            # self.initMainWidget()

            self.callInitScript()

            if not self._loaded:
                return

            # self.bindIface()

    # def cursorDestroyed(self, obj_: Optional[Any] = None) -> None:
    #    """Clean up. Called when cursor has been deleted."""
    #    if not obj_:
    #        obj_ = self.sender()

    #    if not obj_ or obj_ is self.cursor_:
    #        return

    #    del self.cursor_

    """
    Captura evento ocultar


    def hideEvent(self, h):
        pW = self.parentWidget()
        if not pW:
            geo = QtCore.QSize(self.width(), self.height())
            if self.isMinimized():
                geo.setWidth(1)
            elif self.isMaximized():
                geo.setWidth(9999)
        else:
            geo = QtCore.QSize(pW.width(), pW.height())

        #geometry.saveGeometryForm(self.geoName(), geo)
    """

    def show(self) -> None:
        """
        Initialize components of the main widget.

        @param w Widget to initialize. If not set use
        by default the current main widget.
        """
        main_window = application.PROJECT.main_window
        if main_window is None:
            return

        if hasattr(main_window, "_dict_main_widgets"):
            module_name = application.PROJECT.conn_manager.managerModules().activeIdModule()
            if module_name and main_window and module_name in main_window._dict_main_widgets.keys():
                module_window = cast(
                    QtWidgets.QMainWindow, main_window._dict_main_widgets[module_name]
                )

                mdi_area = module_window.centralWidget()
                if isinstance(mdi_area, QtWidgets.QMdiArea):
                    for sub_window in mdi_area.subWindowList():
                        if cast(FLFormDB, sub_window.widget()).formName() == self.formName():
                            mdi_area.setActiveSubWindow(sub_window)
                            return

                    if type(self).__name__ == "FLFormDB":
                        # if not isinstance(self.parent(), QtWidgets.QMdiSubWindow):
                        # size = self.size()
                        mdi_area.addSubWindow(self)

        if self._init_focus_widget is None:
            self._init_focus_widget = self.focusWidget()

        if self._init_focus_widget:
            self._init_focus_widget.setFocus()

        # if not self.tiempo_ini:
        #    self.tiempo_ini = time.time()
        super().show()
        # tiempo_fin = time.time()
        parent_ = self.parent()
        if parent_ and parent_.parent() is None:
            qt_rectangle = self.frameGeometry()
            center_point = self.screen().availableGeometry().center()  # type: ignore [union-attr]
            qt_rectangle.moveCenter(center_point)
            self.move(qt_rectangle.topLeft())

        if not self._showed:  # Prueba al restaurar tabs no inicializan scripts
            self.initScript()

        # if settings.readBoolEntry("application/isDebuggerMode", False):
        #    LOGGER.warning("INFO:: Tiempo de carga de %s: %.3fs %s (iface %s)" %
        #                     (self._action_name, tiempo_fin - self.tiempo_ini, self, self.iface))
        # self.tiempo_ini = None

    def initMainWidget(self, widget: Optional["QtWidgets.QWidget"] = None) -> None:
        """Initialize widget."""

        if widget is not None:
            self.main_widget = widget

        if self.main_widget and not getattr(self.main_widget, "_showed", False):
            self.main_widget.show()

    def child(self, child_name: str) -> Optional["QtWidgets.QWidget"]:
        """Get child by name."""
        ret: Optional["QtCore.QObject"] = None
        try:
            ret = self.findChild(QtWidgets.QWidget, child_name)
        except Exception:
            pass

        if ret is not None:
            from pineboolib.fllegacy import flfielddb, fltabledb

            if isinstance(ret, (flfielddb.FLFieldDB, fltabledb.FLTableDB)):
                if ret._loaded is False:
                    ret.load()

        return cast(QtWidgets.QWidget, ret)

    # def __getattr__(self, name):
    # if getattr(self.script, "form", None):
    #    return getattr(self.script.form, name)
    # else:
    #    qWarning("%s (%s):No se encuentra el atributo %s" % (self, self.iface, name))

    @decorators.not_implemented_warn
    def exportToXml(self, value_: bool) -> None:
        """Export this widget into an xml."""
        from pineboolib.fllegacy.aqsobjects.aqs import AQS

        xml = AQS.toXml(self, True, True)
        print(xml.toString(2))
        pass

    def get_action_widget(self) -> Optional["formdbwidget.FormDBWidget"]:
        """Return main_widget."""
        widget = None
        if self._action.name() in application.PROJECT.actions.keys():
            action = application.PROJECT.actions[self._action.name()]
            widget = (
                action._record_widget
                if self._action_name.startswith("formRecord")
                else action._master_widget
            )

        return widget

    def set_action_widget(self, obj_: "formdbwidget.FormDBWidget"):
        """Set main widget."""

        action = application.PROJECT.actions[self._action.name()]
        if self._action_name.startswith("formRecord"):
            action._record_widget = obj_
        else:
            action._master_widget = obj_

    def get_cursor(self) -> Optional["isqlcursor.ISqlCursor"]:
        """Return action cursor."""
        return application.PROJECT.actions[self._action.name()].cursor()

    def set_cursor(self, cursor: "isqlcursor.ISqlCursor") -> None:
        """Set action cursor."""
        application.PROJECT.actions[self._action.name()].setCursor(cursor)

    @decorators.pyqt_slot()
    @decorators.not_implemented_warn
    def script(self) -> "str":
        """
        Return the script associated with the form.
        """

        return ""

    def get_iface(self) -> Optional["Callable"]:
        """Return script iface."""

        fun = getattr(self.action_widget, "iface", None)
        return fun

    def init_tool_bar(self) -> None:
        """Init bottomtoolbar."""

        self.bottomToolbar.setMinimumSize(self._icon_size)
        hblay = QtWidgets.QHBoxLayout()
        hblay.setContentsMargins(0, 0, 0, 0)
        hblay.setSpacing(0)
        hblay.addStretch()
        self.bottomToolbar.setLayout(hblay)
        self.bottomToolbar.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        self.layout_.addWidget(self.bottomToolbar)

    action_widget = property(get_action_widget, set_action_widget)
    cursor_ = property(get_cursor, set_cursor)  # type: ignore [assignment] # noqa: F821
    iface = property(get_iface)



"""Test_pntablemetadata module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing

from pineboolib.application.metadata import pntablemetadata
from pineboolib import application


class TestCreatePNTableMetaData(unittest.TestCase):
    """TestCreatePNTableMetaData Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test create a PNTableMetaData."""

        from pineboolib.application.metadata import pntablemetadata

        mtd = pntablemetadata.PNTableMetaData("prueba_1", "Alias de prueba_1", "qry_prueba_1")
        mtd_2 = pntablemetadata.PNTableMetaData("prueba_2")

        self.assertEqual(mtd.name(), "prueba_1")
        self.assertEqual(mtd_2.name(), "prueba_2")
        self.assertEqual(mtd.alias(), "Alias de prueba_1")
        self.assertEqual(mtd.isQuery(), True)

        mtd.setName("prueba_2")
        mtd.setAlias("Alias de prueba_2")
        mtd.setQuery("qry_prueba_2")

        self.assertEqual(mtd.query(), "qry_prueba_2")

    def test_basic_2(self) -> None:
        """Test functions."""

        from pineboolib.application.database import pnsqlcursor

        cur = pnsqlcursor.PNSqlCursor("fltest2")
        mtd = cur.metadata()

        self.assertEqual(mtd.fieldType("string_field"), 3)
        self.assertEqual(mtd.fieldType("time_field"), 27)
        self.assertEqual(mtd.fieldType("date_field"), 26)
        self.assertEqual(mtd.fieldType("double_field"), 19)
        self.assertEqual(mtd.fieldType("bool_field"), 18)
        self.assertEqual(mtd.fieldType("uint_field"), 17)
        self.assertEqual(mtd.fieldType("bloqueo"), 200)

        self.assertFalse(mtd.fieldIsCounter("string_field"))
        self.assertTrue(mtd.fieldAllowNull("string_field"))
        self.assertFalse(mtd.fieldAllowNull("bloqueo"))

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()


class TestCopyPNTableMetaData(unittest.TestCase):
    """TestCopyPNTableMetaData Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test copy a PNTableMetaData from other."""

        mtd_1 = application.PROJECT.conn_manager.manager().metadata("flgroups")
        if mtd_1:
            self.assertEqual(mtd_1.alias(), "Grupos de Usuarios")
        mtd_2 = pntablemetadata.PNTableMetaData(mtd_1 or "")

        self.assertEqual(mtd_2.name(), "flgroups")
        self.assertEqual(mtd_2.alias(), "Grupos de Usuarios")
        self.assertEqual(mtd_2.fieldNameToAlias("idgroup"), "Nombre")
        self.assertEqual(mtd_2.primaryKey(), "idgroup")
        self.assertEqual(mtd_2.fieldAliasToName("Nombre"), "idgroup")

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()


class TestPNTableMetaData(unittest.TestCase):
    """TestPNTableMetaData Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test manage a PNTableMetaData."""

        mtd = application.PROJECT.conn_manager.manager().metadata("flgroups")
        mtd_2 = application.PROJECT.conn_manager.manager().metadata("flareas")
        self.assertTrue(mtd is not None)
        self.assertTrue(mtd_2 is not None)
        if mtd is not None and mtd_2 is not None:
            self.assertEqual(mtd.fieldType("descripcion"), 3)
            self.assertEqual(mtd.fieldIsPrimaryKey("descripcion"), False)
            self.assertEqual(mtd.fieldIsPrimaryKey("idgroup"), True)
            self.assertEqual(mtd.fieldIsIndex("descripcion"), 1)
            self.assertEqual(mtd.fieldLength("descripcion"), 100)
            self.assertEqual(mtd.fieldPartInteger("descripcion"), 0)
            self.assertEqual(mtd.fieldPartDecimal("descripcion"), 0)
            self.assertEqual(mtd.fieldCalculated("descripcion"), False)
            self.assertEqual(mtd.fieldVisible("descripcion"), True)

            field = mtd.field("descripcion")
            self.assertTrue(field is not None)
            if field is not None:
                self.assertEqual(field.name(), "descripcion")

                field_list = mtd.fieldList()
                self.assertEqual(len(field_list), 2)

                field_list_2 = mtd.fieldListArray(False)
                self.assertEqual(field_list_2, ["idgroup", "descripcion"])

                field_list_3 = mtd.fieldListArray(True)
                self.assertEqual(field_list_3, ["flgroups.idgroup", "flgroups.descripcion"])

                field_list_4 = mtd.fieldList(True)
                self.assertEqual(field_list_4, ["idgroup", "descripcion"])

            mtd.removeFieldMD("descripcion")
            self.assertEqual(mtd.fieldIsIndex("descripcion"), -1)
            self.assertEqual(mtd.fieldIsUnique("idgroup"), False)

            self.assertEqual(mtd.indexPos("idgroup"), 0)
            self.assertEqual(mtd.fieldNames(), ["idgroup"])
            self.assertEqual(mtd_2.fieldNamesUnlock(), ["bloqueo"])

            self.assertEqual(mtd.concurWarn(), False)
            mtd.setConcurWarn(True)
            self.assertEqual(mtd.concurWarn(), True)

            self.assertEqual(mtd.detectLocks(), False)
            mtd.setDetectLocks(True)
            self.assertEqual(mtd.detectLocks(), True)

            self.assertEqual(mtd.FTSFunction(), "")
            mtd.setFTSFunction("1234_5678")
            self.assertEqual(mtd.FTSFunction(), "1234_5678")

            self.assertEqual(mtd.inCache(), False)
            mtd.setInCache(True)
            self.assertEqual(mtd.inCache(), True)

            field_pos_0 = mtd.indexFieldObject(0)
            self.assertEqual(field_pos_0.name(), "idgroup")

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()


class TestRelationsPNTableMetaData(unittest.TestCase):
    """TestRelationsPNTableMetaData Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test Relations M1 and 1M."""

        mtd_1 = application.PROJECT.conn_manager.manager().metadata("flusers")
        self.assertTrue(mtd_1)
        if mtd_1 is not None:
            self.assertEqual(mtd_1.fieldTableM1("idgroup"), "flgroups")
            self.assertEqual(mtd_1.fieldForeignFieldM1("idgroup"), "idgroup")

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()


class TestCompoundKeyPNTableMetaData(unittest.TestCase):
    """TestCompoundKeyPNTableMetaData Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test CompoundKey."""

        mtd = application.PROJECT.conn_manager.manager().metadata("flseqs")
        self.assertTrue(mtd is not None)
        if mtd is not None:
            field_list = mtd.fieldListOfCompoundKey("campo")
            self.assertTrue(field_list)
            if field_list:
                self.assertEqual(field_list[0].name(), "campo")

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()


class TestCachedFieldsPNTableMetaData(unittest.TestCase):
    """TestRelationsPNTableMetaData Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test cachedFields."""

        mtd_1 = application.PROJECT.conn_manager.manager().metadata("flusers")

        self.assertTrue(mtd_1)
        mtd_1.setCachedFields("*")
        if mtd_1 is not None:
            self.assertTrue(mtd_1.useCachedFields())
            self.assertEqual(mtd_1.cachedFields(), "*")
        else:
            raise Exception("mtd_1 is None")

        mtd_2 = application.PROJECT.conn_manager.manager().metadata("flusers")

        self.assertTrue(mtd_2)
        mtd_2.setCachedFields("iduser,description")
        if mtd_2 is not None:
            self.assertTrue(mtd_2.useCachedFields())
            self.assertEqual(mtd_2.cachedFields(), "iduser,description")
        else:
            raise Exception("mtd_2 is None")

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



# -*- coding: utf-8 -*-
"""
Collection of controls used on Pineboo.

Those are loaded from selected DGI.
"""
from PyQt6 import QtCore, QtWidgets, QtGui, QtXml  # type: ignore[import] # noqa: F401

from typing import Any

from pineboolib.core.system import System  # noqa: F401

from pineboolib.qsa.utils import MathClass, NumberAttr, user_id, session, require  # noqa: F401
from pineboolib import application
from pineboolib.application.packager.pnunpacker import PNUnpacker as AQUnpacker  # noqa: F401
from pineboolib.application.packager.pnpackager import PNPackager as AQPackager  # noqa: F401

from pineboolib.fllegacy.systype import SysType

from pineboolib.q3widgets.qcombobox import QComboBox  # noqa: F401
from pineboolib.q3widgets.qtable import QTable  # noqa: F401
from pineboolib.q3widgets.qlayoutwidget import QLayoutWidget  # noqa: F401
from pineboolib.q3widgets.qtoolbutton import QToolButton  # noqa: F401
from pineboolib.q3widgets.qtabwidget import QTabWidget  # noqa: F401
from pineboolib.q3widgets.qlabel import QLabel  # noqa: F401
from pineboolib.q3widgets.qgroupbox import QGroupBox  # noqa: F401
from pineboolib.q3widgets.qlistview import QListView  # noqa: F401
from pineboolib.q3widgets.qpushbutton import QPushButton  # noqa: F401
from pineboolib.q3widgets.qtextedit import QTextEdit  # noqa: F401
from pineboolib.q3widgets.qlineedit import QLineEdit  # noqa: F401
from pineboolib.q3widgets.qdateedit import QDateEdit  # noqa: F401
from pineboolib.q3widgets.qtimeedit import QTimeEdit  # noqa: F401
from pineboolib.q3widgets.qcheckbox import QCheckBox  # noqa: F401
from pineboolib.q3widgets.qwidget import QWidget  # noqa: F401

# from pineboolib.q3widgets.messagebox import QMessageBox  # noqa: F401
from pineboolib.q3widgets.qbuttongroup import QButtonGroup  # noqa: F401
from pineboolib.q3widgets.qdialog import QDialog  # noqa: F401
from pineboolib.q3widgets.qvboxlayout import QVBoxLayout  # noqa: F401
from pineboolib.q3widgets.qhboxlayout import QHBoxLayout  # noqa: F401
from pineboolib.q3widgets.qframe import QFrame  # noqa: F401
from pineboolib.q3widgets.qmainwindow import QMainWindow  # noqa: F401
from pineboolib.q3widgets.qmenu import QMenu  # noqa: F401
from pineboolib.q3widgets.qtoolbar import QToolBar  # noqa: F401
from pineboolib.q3widgets.qaction import QAction  # noqa: F401
from pineboolib.q3widgets.qdataview import QDataView  # noqa: F401
from pineboolib.q3widgets.qbytearray import QByteArray  # noqa: F401
from pineboolib.q3widgets.qradiobutton import QRadioButton  # noqa: F401
from pineboolib.q3widgets.qspinbox import QSpinBox  # noqa: F401
from pineboolib.q3widgets.qtextstream import QTextStream  # noqa: F401
from pineboolib.q3widgets.qpopupmenu import QPopupMenu  # noqa: F401
from pineboolib.q3widgets.qiconset import QIconSet  # noqa: F401
from pineboolib.q3widgets.qhbuttongroup import QHButtonGroup  # noqa: F401
from pineboolib.q3widgets.qvbuttongroup import QVButtonGroup  # noqa: F401
from pineboolib.q3widgets.qobject import QObject  # noqa: F401

from pineboolib.q3widgets.qeventloop import QEventLoop  # noqa: F401
from pineboolib.q3widgets.qlistviewwidget import QListViewWidget  # noqa: F401
from pineboolib.q3widgets.qdatetime import QDateTime  # noqa: F401
from pineboolib.q3widgets.qdir import QDir  # noqa: F401

from pineboolib.q3widgets.qhttp import QHttp, QHttpResponseHeader, QHttpRequestHeader  # noqa: F401

from PyQt6.QtGui import QActionGroup  # type: ignore[import] # noqa: F401
from PyQt6.QtWidgets import QInputDialog  # type: ignore[import] # noqa: F401
from PyQt6.QtWidgets import QApplication  # noqa: F401


from PyQt6.QtWidgets import QStyleFactory  # noqa: F401
from PyQt6.QtWidgets import QFontDialog  # noqa: F401
from PyQt6.QtWidgets import QDockWidget  # noqa: F401
from PyQt6.QtWidgets import QMdiSubWindow  # noqa: F401
from PyQt6.QtWidgets import QSizePolicy  # noqa: F401
from PyQt6.QtWidgets import QToolBox  # noqa: F401
from PyQt6.QtWidgets import QProgressDialog  # noqa: F401
from PyQt6.QtWidgets import QFileDialog  # noqa: F401
from PyQt6.QtWidgets import QTreeWidget  # noqa: F401
from PyQt6.QtWidgets import QTreeWidgetItem  # noqa: F401
from PyQt6.QtWidgets import QTreeWidgetItemIterator  # noqa: F401
from PyQt6.QtWidgets import QListWidgetItem  # noqa: F401
from PyQt6.QtWidgets import QMdiArea  # noqa: F401
from PyQt6.QtWidgets import QMessageBox  # noqa: F401

from PyQt6.QtCore import QSignalMapper  # type: ignore[import] # noqa: F401
from PyQt6.QtCore import QSize  # noqa: F401

from PyQt6.QtCore import QBuffer  # noqa: F401

from PyQt6.QtGui import QPainter  # noqa: F401
from PyQt6.QtGui import QBrush  # noqa: F401
from PyQt6.QtGui import QKeySequence  # noqa: F401
from PyQt6.QtGui import QPixmap  # noqa: F401
from PyQt6.QtGui import QImage  # noqa: F401
from PyQt6.QtGui import QIcon  # noqa: F401
from PyQt6.QtGui import QColor  # noqa: F401


from PyQt6.QtXml import QDomDocument  # type: ignore[import] # noqa: F401

# Clases FL
from PyQt6.QtXml import QDomDocument as FLDomDocument  # noqa: F401
from PyQt6.QtXml import QDomElement as FLDomElement  # noqa: F401
from PyQt6.QtXml import QDomNode as FLDomNode  # noqa: F401
from PyQt6.QtXml import QDomNodeList as FLDomNodeList  # noqa: F401

from pineboolib.qsa.formdbwidget import FormDBWidget  # noqa: F401
from pineboolib.qsa.object_class import ObjectClass  # noqa: F401

from pineboolib.fllegacy.fltable import FLTable  # noqa: F401
from pineboolib.fllegacy.fllineedit import FLLineEdit  # noqa: F401
from pineboolib.fllegacy.fltimeedit import FLTimeEdit  # noqa: F401
from pineboolib.fllegacy.fldateedit import FLDateEdit  # noqa: F401
from pineboolib.fllegacy.flpixmapview import FLPixmapView  # noqa: F401
from pineboolib.fllegacy.fllistviewitem import FLListViewItem  # noqa: F401
from pineboolib.fllegacy.fldatatable import FLDataTable  # noqa: F401
from pineboolib.fllegacy.flcheckbox import FLCheckBox  # noqa: F401
from pineboolib.application.pnapplication import TextEditOutput as FLTextEditOutput  # noqa: F401
from pineboolib.fllegacy.flspinbox import FLSpinBox  # noqa: F401
from pineboolib.fllegacy.fltabledb import FLTableDB  # noqa: F401
from pineboolib.fllegacy.flfielddb import FLFieldDB  # noqa: F401
from pineboolib.fllegacy.flformdb import FLFormDB  # noqa: F401
from pineboolib.fllegacy.flformrecorddb import FLFormRecordDB  # noqa: F401
from pineboolib.fllegacy.flformsearchdb import FLFormSearchDB  # noqa: F401
from pineboolib.fllegacy.fldoublevalidator import FLDoubleValidator  # noqa: F401
from pineboolib.fllegacy.flintvalidator import FLIntValidator  # noqa: F401
from pineboolib.fllegacy.fluintvalidator import FLUIntValidator  # noqa: F401
from pineboolib.fllegacy.flcodbar import FLCodBar  # noqa: F401
from pineboolib.fllegacy.flwidget import FLWidget  # noqa: F401
from pineboolib.fllegacy.flworkspace import FLWorkSpace  # noqa: F401
from pineboolib.fllegacy.flutil import FLUtil  # noqa: F401
from pineboolib.fllegacy.flsettings import FLSettings  # noqa: F401
from pineboolib.fllegacy.flposprinter import FLPosPrinter  # noqa: F401
from pineboolib.fllegacy.flsqlquery import FLSqlQuery  # noqa: F401
from pineboolib.fllegacy.flserialport import FLSerialPort  # noqa: F401
from pineboolib.fllegacy.flsqlcursor import FLSqlCursor  # noqa: F401
from pineboolib.fllegacy.flnetwork import FLNetwork  # noqa: F401
from pineboolib.fllegacy.flreportviewer import FLReportViewer  # noqa: F401
from pineboolib.fllegacy.flreportengine import FLReportEngine  # noqa: F401
from pineboolib.fllegacy.fljasperengine import FLJasperEngine  # noqa: F401
from pineboolib.fllegacy.fljasperviewer import FLJasperViewer  # noqa: F401
from pineboolib.fllegacy.flfastcgi import FLFastCgi  # noqa: F401

from pineboolib.fllegacy.flapplication import FLApplication  # noqa: F401
from pineboolib.fllegacy.flvar import FLVar  # noqa: F401
from pineboolib.fllegacy.flsmtpclient import FLSmtpClient  # noqa: F401
from pineboolib.fllegacy.flscripteditor import FLScriptEditor  # noqa: F401

# Clases QSA

from PyQt6.QtGui import QColor as Color  # noqa: F401
from PyQt6.QtWidgets import QLabel as Label  # noqa: F401
from PyQt6.QtGui import QFont as Font  # noqa: F401


from pineboolib.q3widgets.line import Line  # noqa: F401
from pineboolib.q3widgets.checkbox import CheckBox  # noqa: F401
from pineboolib.q3widgets.combobox import ComboBox  # noqa: F401
from pineboolib.q3widgets.textedit import TextEdit  # noqa: F401
from pineboolib.q3widgets.lineedit import LineEdit  # noqa: F401
from pineboolib.q3widgets.messagebox import MessageBox  # noqa: F401
from pineboolib.q3widgets.radiobutton import RadioButton  # noqa: F401
from pineboolib.q3widgets.filedialog import FileDialog  # noqa: F401
from pineboolib.q3widgets.spinbox import SpinBox  # noqa: F401

from pineboolib.q3widgets.dialog import Dialog  # noqa: F401
from pineboolib.q3widgets.groupbox import GroupBox  # noqa: F401

from pineboolib.q3widgets.numberedit import NumberEdit  # noqa: F401
from pineboolib.q3widgets.dateedit import DateEdit  # noqa: F401
from pineboolib.q3widgets.timeedit import TimeEdit  # noqa: F401
from pineboolib.q3widgets.picture import Picture  # noqa: F401
from pineboolib.q3widgets.rect import Rect  # noqa: F401
from pineboolib.q3widgets.size import Size  # noqa: F401
from pineboolib.q3widgets.pixmap import Pixmap  # noqa: F401

from pineboolib.fllegacy.aqsobjects.aqsettings import AQSettings  # noqa: F401
from pineboolib.fllegacy.aqsobjects.aqsqlquery import AQSqlQuery  # noqa: F401
from pineboolib.fllegacy.aqsobjects.aqsqlcursor import AQSqlCursor  # noqa: F401
from pineboolib.fllegacy.aqsobjects.aqutil import AQUtil  # noqa: F401
from pineboolib.fllegacy.aqsobjects.aqsql import AQSql  # noqa: F401
from pineboolib.fllegacy.aqsobjects.aqsmtpclient import AQSmtpClient  # noqa: F401
from pineboolib.fllegacy.aqsobjects.aqs import AQS  # noqa: F401
from pineboolib.fllegacy.aqsobjects.aqsignalmapper import AQSignalMapper  # noqa: F401
from pineboolib.fllegacy.aqsobjects.aqssproject import AQSSProject  # noqa: F401
from pineboolib.fllegacy.aqsobjects.aqobjectquerylist import (  # noqa: F401
    aq_object_query_list as AQObjectQueryList,
)


from pineboolib.core.utils.utils_base import is_deployed as __is_deployed
from pineboolib.core.utils.utils_base import is_library as __is_library
from pineboolib.application.database.orm.utils import OrmManager
from pineboolib.application.database.utils import ClassManager

from pineboolib.application.qsadictmodules import from_project, orm_  # noqa: F401


from pineboolib.fllegacy.aqsobjects.aqboolflagstate import AQBoolFlagState  # noqa: F401
from pineboolib.fllegacy.aqsobjects.aqboolflagstate import AQBoolFlagStateList  # noqa: F401
from pineboolib.fllegacy.aqsobjects.aqsbuttongroup import AQSButtonGroup  # noqa: F401

from pineboolib.application.signatures.xml_digest import XmlDigest  # noqa: F401
from pineboolib.application.signatures.pdf_digest import PdfDigest  # noqa: F401
from pineboolib.application.signatures.pdf_qr import PdfQr  # noqa: F401

if not __is_deployed() and not __is_library():
    from pineboolib.fllegacy.aqsobjects.aqods import AQOdsGenerator, AQOdsSpreadSheet  # noqa: F401
    from pineboolib.fllegacy.aqsobjects.aqods import AQOdsSheet, AQOdsRow  # noqa: F401
    from pineboolib.fllegacy.aqsobjects.aqods import AQOdsStyle  # noqa: F401
    from pineboolib.fllegacy.aqsobjects.aqods import AQOdsImage  # noqa: F401
    from pineboolib.fllegacy.aqsobjects.aqods import aq_ods_color as AQOdsColor  # noqa: F401

qApp = QApplication  # noqa: F401
ORM_MANAGER = OrmManager()
CLASS_MANAGER = ClassManager()

QSA_SYS = SysType()

QS_PROJECT = AQSSProject()
MATH = MathClass()
NUMBER_ATT = NumberAttr()
AQ_APP = application.PROJECT.aq_app


class Application:
    """
    Emulate QS Application class.

    The "Data" module uses "Application.formRecorddat_processes" to read the module.
    """

    def __getattr__(self, name: str) -> Any:
        """Emulate any method and retrieve application action module specified."""

        from_project(name)



"""Test_pnparameterquery module."""

import unittest
from pineboolib.application.database import pnparameterquery


class TestPNParameterQuery(unittest.TestCase):
    """TestPNParameterQuery Class."""

    def test_full(self) -> None:
        """Test full"""

        param = pnparameterquery.PNParameterQuery("prueba", "Campo de prueba", 2)
        param.setValue("Esto es una prueba")

        self.assertEqual(param.name(), "prueba")
        self.assertEqual(param.alias(), "Campo de prueba")
        self.assertEqual(param.type(), 2)
        self.assertEqual(param.value(), "Esto es una prueba")



"""Input module."""

from PyQt6 import QtWidgets  # type: ignore[import]
from typing import Any, Optional, Union, Iterable


class Input(object):
    """
    Data entry dialog.
    """

    @classmethod
    def getText(cls, question: str, prevtxt: str = "", title: str = "Pineboo") -> Optional[str]:
        """
        Return Text.

        @param question. Label of the dialogue.
        @param prevtxt. Initial value to specify in the field.
        @param title. Title of the dialogue.
        @return string of collected text.
        """
        parent = QtWidgets.QApplication.activeWindow()
        text, result = QtWidgets.QInputDialog.getText(
            parent, title, question, QtWidgets.QLineEdit.EchoMode.Normal, prevtxt
        )
        if not result:
            return None
        return text

    @classmethod
    def getNumber(
        cls, question: str, value: Union[str, float], part_decimal: int = 0, title: str = "Pineboo"
    ) -> float:
        """Return number."""

        parent = QtWidgets.QApplication.activeWindow()
        text, result = QtWidgets.QInputDialog.getText(
            parent,
            title,
            question,
            QtWidgets.QLineEdit.EchoMode.Normal,
            str(round(float(value), part_decimal)),
        )
        ret: float = 0.00
        if result:
            ret = float(text)

        return ret

    @classmethod
    def getItem(
        cls, question: str, items_list: Iterable = [], editable: bool = True, title: str = "Pineboo"
    ) -> Any:
        """
        Return Item.

        @param question. Label of the dialogue.
        @param item_list. Items List.
        @param title. Title of the dialogue.
        @return item, Selected item.
        """
        parent = QtWidgets.QApplication.activeWindow()
        items = []
        for i in items_list:
            items.append(i)

        text, result = QtWidgets.QInputDialog.getItem(parent, title, question, items, 0, editable)
        if not result:
            return None
        return text



"""
Tests for loader.main.
"""
import unittest
from unittest.mock import patch, Mock
from pineboolib.loader.options import parse_options
from pineboolib.loader import main


class TestMain(unittest.TestCase):
    """Test Main load."""

    @patch("pineboolib.loader.main.sys.exit")
    @patch("pineboolib.loader.main.exec_main")
    @patch("pineboolib.loader.options.parse_options")
    def test_startup(
        self, mock_parse_options: Mock, mock_exec_main: Mock, mock_sys_exit: Mock
    ) -> None:
        """Test bug where logging tries to get incorrect options."""
        options = parse_options(custom_argv=[])
        mock_parse_options.return_value = options
        main.startup()
        mock_parse_options.assert_called_once()
        mock_exec_main.assert_called_once()
        mock_sys_exit.assert_called_once()



# -*- coding: utf-8 -*-
"""
PNAccessControlList Module.

Manage access lists to limit the application to users..
"""
from PyQt6 import QtCore, QtXml  # type: ignore[import]

from pineboolib import application

from pineboolib.application.database import pnsqlquery
from pineboolib.application.metadata import pntablemetadata
from pineboolib.application.acls import pnaccesscontrolfactory

from pineboolib import logging
from typing import Dict, Optional, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.application.acls import pnaccesscontrol  # pragma: no cover
    from PyQt6 import QtWidgets  # pragma: no cover


LOGGER: "logging.Logger" = logging.get_logger(__name__)


class PNAccessControlLists(object):
    """PNAccessControlList Class."""

    """
    Nombre que identifica la lista de control de acceso actualmente establecida.

    Generalmente corresponderá con el identificador del registro de la tabla "flacls" que se utilizó para crear "acl.xml".
    """
    _name: str

    """
    Diccionario (lista) que mantiene los objetos de las reglas de control de acceso establecidas.
    La clave que identifica a cada objeto está formada por el siguiente literal:

    \\code

    PNAccessControl::type + "::" + PNAccessControl::name + "::" + PNAccessControl::user

    \\endcode
    """

    _access_control_list: Dict[str, "pnaccesscontrol.PNAccessControl"]

    def __init__(self) -> None:
        """Initialize the class."""

        self._name = ""
        self._access_control_list = {}

    def __del__(self) -> None:
        """Process when destroying the class."""

        self._access_control_list.clear()
        del self._access_control_list

    def name(self) -> Optional[str]:
        """
        Return the name that identifies the currently established access control list.

        @return Name the current access control list.
        """
        return self._name

    def init(self, _acl_xml: str = "") -> None:
        """
        Read the file "acl.xml" and establish a new access control list.

        If the file "acl.xml" cannot be read, the access control list is empty and
        no access control will be processed on any object.

        @param _acl_xml XML content with the definition of the access control list.
        """
        if not application.ENABLE_ACLS:
            LOGGER.warning("Init canceled.ACLS usage is disabled")
            return

        if not _acl_xml:
            _acl_xml = application.PROJECT.conn_manager.managerModules().content("acl.xml")

        doc = QtXml.QDomDocument("ACL")

        self._access_control_list.clear()

        if _acl_xml and not doc.setContent(_acl_xml):
            LOGGER.warning(QtCore.QObject().tr("Lista de control de acceso errónea"))  # type: ignore [unreachable]
            return

        doc_elem = doc.documentElement()
        node = doc_elem.firstChild()

        if node.isNull():
            return

        while not node.isNull():
            element = node.toElement()
            if element is not None:
                tag_name = element.tagName()
                if tag_name == "name":
                    self._name = element.text()
                    node = node.nextSibling()
                    continue

                rule = pnaccesscontrolfactory.PNAccessControlFactory().create(tag_name)
                rule.set(element)
                self._access_control_list[
                    "%s::%s::%s" % (rule.type(), rule.name(), rule.user())
                ] = rule

            node = node.nextSibling()

        LOGGER.warning(QtCore.QObject().tr("Lista de control de acceso cargada"))

    def process(
        self, obj: Optional[Union["QtWidgets.QWidget", "pntablemetadata.PNTableMetaData"]] = None
    ) -> None:
        """
        Process a high-level object according to the established access control list.

        @param obj High-level object to which access control is applied. It must be or inherit from the QObject class.
        """

        if obj is None or not self._access_control_list:
            return None

        type_: str = pnaccesscontrolfactory.PNAccessControlFactory().type(obj)

        name_: str = (
            obj.name() if hasattr(obj, "name") else obj.objectName()  # type: ignore [union-attr]
        )  # type: ignore [union-attr]

        user_: str = application.PROJECT.conn_manager.mainConn().user()

        if "" in (type_, name_, user_):
            return None

        keys_ = []
        keys_.append("%s::%s::%s" % (type_, name_, user_))
        if type_ == "mainwindow":
            keys_.append("%s::%s::%s" % (type_, "container", user_))

        for key in keys_:
            if key in self._access_control_list.keys():
                self._access_control_list[key].processObject(obj)

    def install_acl(self, idacl: str) -> None:
        """
        Create a new file "acl.xml" and store it replacing the previous one, if it exists.

        @param idacl Record identifier of the "flacls" table to use to create "acl.xml".
        """
        doc = QtXml.QDomDocument("ACL")

        root = doc.createElement("ACL")
        doc.appendChild(root)

        name = doc.createElement("name")
        root.appendChild(name)
        text_node = doc.createTextNode(idacl)
        name.appendChild(text_node)

        qry = pnsqlquery.PNSqlQuery()

        qry.setTablesList("flacs")
        qry.setSelect("idac,tipo,nombre,iduser,idgroup,degrupo,permiso")
        qry.setFrom("flacs")
        qry.setWhere("idacl='%s'" % idacl)
        qry.setOrderBy("prioridad DESC, tipo")
        qry.setForwardOnly(True)

        if qry.exec_():
            # step = 0
            # progress = util.ProgressDialog(util.tr("Instalando control de acceso..."), None, q.size(), None, None, True)
            # progress.setCaption(util.tr("Instalando ACL"))
            # progress.setMinimumDuration(0)
            # progress.setProgress(++step)
            while qry.next():
                self.make_rule(qry, doc)
                # progress.setProgress(++step)

            application.PROJECT.conn_manager.managerModules().setContent(
                "acl.xml", "sys", doc.toString()
            )

    def make_rule(
        self,
        qry: Optional["pnsqlquery.PNSqlQuery"] = None,
        dom_document: Optional["QtXml.QDomDocument"] = None,
    ) -> None:
        """
        Create the corresponding DOM node (s) to a record in the "flacs" table.

        Use PNAccessControlLists :: makeRuleUser or PNAccessControlLists :: makeRuleGroup depending on whether the registry
        to which the query points indicates that the rule is for a user or a group. If the record indicates a
        user will create a user rule, if you indicate a group a user rule will be created for each of
        Group member users.

        @param q Query about the "flacs" table positioned in the register to be used to construct the rule (s).
        @param d DOM / XML document in which you will insert the node (s) that describe the access control rule (s).
        """
        if not qry:
            return None

        if not dom_document:
            return None

        self.make_rule_group(qry, dom_document, str(qry.value(4))) if qry.value(
            5
        ) else self.make_rule_user(qry, dom_document, str(qry.value(3)))

    def make_rule_user(
        self,
        qry: Optional["pnsqlquery.PNSqlQuery"] = None,
        dom_document: Optional["QtXml.QDomDocument"] = None,
        iduser: str = "",
    ) -> None:
        """
        Create a DOM node corresponding to a record in the "flacs" table and for a given user.

        @param q Query about the "flacs" table positioned in the register to be used to construct the rule.
        @param d DOM / XML document in which you will insert the node that describes the access control rule.
        @param iduser Identifier of the user used in the access control rule.
        """
        if not iduser:
            return None

        if not qry:
            return None
        if not dom_document:
            return None

        rule: "pnaccesscontrol.PNAccessControl" = (
            pnaccesscontrolfactory.PNAccessControlFactory().create(str(qry.value(1)))
        )

        rule.setName(str(qry.value(2)))
        rule.setUser(iduser)
        rule.setPerm(str(qry.value(6)))

        qry_acos = pnsqlquery.PNSqlQuery()
        qry_acos.setTablesList("flacos")
        qry_acos.setSelect("nombre,permiso")
        qry_acos.setFrom("flacos")
        qry_acos.setWhere("idac ='%s'" % qry.value(0))
        qry_acos.setForwardOnly(True)

        acos = []

        if qry_acos.exec_():
            while qry_acos.next():
                acos.append(str(qry_acos.value(0)))
                acos.append((qry_acos.value(1)))

        rule.setAcos(acos)
        rule.get(dom_document)

    def make_rule_group(
        self, qry: pnsqlquery.PNSqlQuery, dom_document: QtXml.QDomDocument, idgroup: str = ""
    ) -> None:
        """
        Create several DOM nodes corresponding to a record in the "flacs" table and for a specific user group.

        The function of this method is to create a rule for each of the group member users, using
        PNAccessControlLists :: makeRuleUser.

        @param q Query about the "flacs" table positioned in the register to use to build the rules.
        @param d DOM / XML document in which the nodes that describe the access control rules will be inserted.
        @param idgroup Identifier of the user group.
        """
        if not idgroup or not qry or not dom_document:
            return

        qry_users = pnsqlquery.PNSqlQuery()

        qry_users.setTablesList("flusers")
        qry_users.setSelect("iduser")
        qry_users.setFrom("flusers")
        qry_users.setWhere("idgroup='%s'" % idgroup)
        qry_users.setForwardOnly(True)

        if qry_users.exec_():
            while qry_users.next():
                self.make_rule_user(qry, dom_document, str(qry_users.value(0)))



"""Test flmodules module."""


import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib import logging, application
from pineboolib.system_module.scripts.tests import fixture_path
from pineboolib.core.utils import utils_base
import codecs

LOGGER = logging.get_logger("eneboo_%s" % __name__)


class TestFlModules(unittest.TestCase):
    """TestFlModules class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        utils_base.FORCE_DESKTOP = True
        init_testing()

    def test_cargar_ficheros(self) -> None:
        """Test load files into Database."""
        from pineboolib.qsa import qsa
        from pineboolib.plugins.mainform import eneboo

        application.PROJECT.main_window = eneboo.MainForm()

        cur_area = qsa.FLSqlCursor("flareas")
        cur_area.select()
        cur_area.setModeAccess(cur_area.Insert)
        cur_area.refreshBuffer()
        cur_area.setValueBuffer("idarea", "sysco")
        cur_area.setValueBuffer("descripcion", "area sysco")
        self.assertTrue(cur_area.commitBuffer())

        cur_mod = qsa.FLSqlCursor("flmodules")
        cur_mod.select()
        cur_mod.setModeAccess(cur_mod.Insert)
        cur_mod.refreshBuffer()
        cur_mod.setValueBuffer("idarea", "sysco")
        cur_mod.setValueBuffer("idmodulo", "mod_sysco")
        cur_mod.setValueBuffer("version", "1")
        cur_mod.setValueBuffer("descripcion", "modulo mod_sysco")
        self.assertTrue(cur_mod.commitBuffer())

        # qsa.from_project("formRecordflmodules").cargarFicheros(fixture_path("scripts"), "*.py")
        cursor = qsa.FLSqlCursor("flmodules")
        cursor.insertRecord(False)
        form = qsa.from_project("formRecordflmodules").form
        cursor_form = form.cursor()
        self.assertEqual(cursor, cursor_form)
        cursor_form.setValueBuffer("idmodulo", "mod_sysco")
        cursor_form.setValueBuffer("idarea", "sysco")
        # cursor_form.commitBuffer()
        qsa.from_project("formRecordflmodules").load_files(fixture_path("scripts"), "*.py")
        qsa.from_project("formRecordflmodules").load_files(fixture_path("scripts"), "*.qs")
        self.assertTrue(
            qsa.sys.disableObj(qsa.from_project("formRecordflmodules").form, "toolButtonEdit")
        )

        cursor_form.commit()

        self.assertTrue(
            qsa.sys.setObjText(
                qsa.from_project("formRecordflmodules").form, "flfielddb_5", "prueba"
            )
        )
        self.assertTrue(
            qsa.from_project("formRecordflmodules").child("flfielddb_5").value(), "prueba"
        )
        self.assertTrue(
            qsa.sys.filterObj(qsa.from_project("formRecordflmodules").form, "flfielddb_5", "prueba")
        )
        self.assertTrue(
            qsa.sys.testAndRun(qsa.from_project("formRecordflmodules").form, "flfielddb_5")
        )

        qry = qsa.FLSqlQuery()
        qry.setSelect("contenido")
        qry.setFrom("flfiles")
        qry.setWhere("nombre = 'prueba.py'")
        qry.exec_()
        data_qry_py = ""
        if qry.first():
            data_qry_py = qry.value(0)

        qry2 = qsa.FLSqlQuery()
        qry2.setSelect("contenido")
        qry2.setFrom("flfiles")
        qry2.setWhere("nombre = 'prueba.qs'")
        qry2.exec_()
        data_qry_qs = ""
        if qry2.first():
            data_qry_qs = qry2.value(0)

        file_py = codecs.open(fixture_path("scripts/prueba.py"), "r", encoding="UTF8")
        data_file_py = file_py.read()
        file_py.close()

        file_qs = codecs.open(fixture_path("scripts/prueba.qs"), "r", encoding="ISO-8859-1")
        data_file_qs = file_qs.read()
        file_qs.close()

        file_qs_bad = codecs.open(
            fixture_path("scripts/prueba.qs"), "r", encoding="UTF8", errors="ignore"
        )
        data_file_qs_bad = file_qs_bad.read()
        file_qs_bad.close()

        self.assertEqual(data_file_py, data_qry_py)
        self.assertEqual(data_file_qs, data_qry_qs)
        self.assertNotEqual(data_file_qs_bad, data_qry_qs)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""Flcheckbox module."""
# -*- coding: utf-8 -*-

from pineboolib.q3widgets import qcheckbox
from PyQt6 import QtWidgets  # type: ignore[import]
from typing import Optional


class FLCheckBox(qcheckbox.QCheckBox):
    """FLCheckBox class."""

    def __init__(
        self, parent: Optional["QtWidgets.QWidget"] = None, num_rows: Optional[int] = None
    ) -> None:
        """Inicialize."""
        super().__init__(parent)
        self.setContentsMargins(0, 0, 0, 0)



"""Test_proxy module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing

from pineboolib.qsa import qsa


class TestProxy(unittest.TestCase):
    """TestProcess Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_duplicate(self) -> None:
        """Test Duplicate."""

        module_1 = qsa.from_project("formflareas")
        module_2 = qsa.from_project("formflareas")
        self.assertEqual(module_1.iface, module_2.iface)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""IMainWindow module."""

from PyQt6 import QtWidgets  # type: ignore[import]

from typing import List, TYPE_CHECKING, Optional, Dict

if TYPE_CHECKING:
    from pineboolib.fllegacy import flworkspace  # noqa: F401 # pragma: no cover


class IMainWindow(QtWidgets.QMainWindow):
    """IMainWindow class."""

    _p_work_space: Optional["flworkspace.FLWorkSpace"]
    initialized_mods_: List[str]
    _dict_main_widgets: Dict[str, "QtWidgets.QWidget"]
    tab_widget: "QtWidgets.QTabWidget"
    container_: Optional["QtWidgets.QMainWindow"]
    main_widget: "QtWidgets.QWidget"

    def __init__(self):
        """Initialize."""
        super().__init__()
        self._dict_main_widgets = {}

    def writeState(self) -> None:
        """Write settings back to disk."""
        pass  # pragma: no cover

    def readState(self) -> None:
        """Read settings."""

        pass  # pragma: no cover

    def createUi(self, ui_file: str) -> None:
        """Create UI from a file."""

        pass  # pragma: no cover

    def writeStateModule(self) -> None:
        """Write settings for modules."""

        pass  # pragma: no cover

    def readStateModule(self) -> None:
        """Read settings for module."""

        pass  # pragma: no cover

    def initScript(self) -> None:
        """Startup process."""

        pass  # pragma: no cover

    def reinitScript(self):
        """Reinit script."""

        pass  # pragma: no cover

    def loadTabs(self) -> None:
        """Load tabs."""

        pass  # pragma: no cover

    def initToolBar(self) -> None:
        """Initialize toolbar."""

        pass  # pragma: no cover

    def initMenuBar(self) -> None:
        """Initialize menus."""

        pass  # pragma: no cover

    def windowMenuAboutToShow(self) -> None:
        """Signal called before window menu is shown."""

        pass  # pragma: no cover

    def activateModule(self, idm=None) -> None:
        """Initialize module."""

        pass  # pragma: no cover

    def existFormInMDI(self, form_name: str) -> bool:
        """Return if named FLFormDB is open."""

        return True  # pragma: no cover

    def windowMenuActivated(self, id_window: int) -> None:
        """Signal called when user clicks on menu."""

        pass  # pragma: no cover

    def windowClose(self) -> None:
        """Signal called on close."""

        pass  # pragma: no cover

    def toggleToolBar(self, toggle: bool) -> None:
        """Show or hide toolbar."""

        pass  # pragma: no cover

    def toggleStatusBar(self, toggle: bool) -> None:
        """Toggle status bar."""

        pass  # pragma: no cover

    def initToolBox(self) -> None:
        """Initialize toolbox."""

        pass  # pragma: no cover

    def setCaptionMainWidget(self, value: str) -> None:
        """Set application title."""

        pass  # pragma: no cover

    def setMainWidget(self, widget: "QtWidgets.QWidget") -> None:
        """Set mainWidget."""

        pass  # pragma: no cover



"""Filedialog module."""
# -*- coding: utf-8 -*-
from PyQt6 import QtWidgets  # type: ignore[import]

from pineboolib.core import system as core_system
import os
from typing import List


class FileDialog(object):
    """FileDialog class."""

    @staticmethod
    def getOpenFileName(filter: str = "*", caption: str = "Pineboo") -> str:
        """Show a dialog to choose a file."""

        folder = core_system.System.getenv("HOME")

        file_list = QtWidgets.QFileDialog.getOpenFileName(None, caption, folder, filter)
        return file_list[0] if file_list else ""

    @staticmethod
    def getOpenFileNames(
        folder: str = ".", filter: str = "*", caption: str = "Pineboo"
    ) -> List[str]:
        """Show a dialog to choose a file."""

        obj = QtWidgets.QFileDialog.getOpenFileNames(None, caption, folder, filter)
        return obj[0] if obj else []

    @staticmethod
    def getSaveFileName(filter: str = "*", caption: str = "Pineboo") -> str:
        """Show a dialog to save a file."""

        folder = core_system.System.getenv("HOME")

        ret = QtWidgets.QFileDialog.getSaveFileName(None, caption, folder, filter)
        return ret[0] if ret else ""

    @staticmethod
    def getExistingDirectory(folder: str = "", caption: str = "Pineboo") -> str:
        """Show a dialog to choose a directory."""

        if not os.path.exists(folder):
            folder = "%s/" % core_system.System.getenv("HOME")

        ret = QtWidgets.QFileDialog.getExistingDirectory(
            None, caption, folder, QtWidgets.QFileDialog.Option.ShowDirsOnly
        )
        return "%s/" % ret if ret else ""



"""Picture module."""
from PyQt6 import QtGui  # type: ignore[import]

from typing import Callable


class Picture(QtGui.QPicture):
    """Picture class."""

    def __getattr__(self, name: str) -> Callable:
        """Return painter attributes."""

        painter = QtGui.QPainter(self)
        ret = getattr(painter, name, None)
        if ret is None:
            raise AttributeError("Attribute %s not found!" % name)

        return ret



"""Test query values module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.qsa import qsa


class TestQueryValues(unittest.TestCase):
    """TestQueryValues class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_pnsqlcursor(self) -> None:
        """Test PNSqlCursor received values."""
        cur = qsa.FLSqlCursor("fltest")
        cur.setModeAccess(cur.Insert)
        cur.refreshBuffer()
        cur.setValueBuffer("date_field", "2000-01-01")
        self.assertTrue(cur.commitBuffer())
        self.assertTrue(isinstance(cur.valueBuffer("date_field"), qsa.Date))

    def test_pnsqlcursor_fake(self) -> None:
        """Test PNSqlCursor_fake received values."""
        class_fltest2 = qsa.orm.fltest
        self.assertTrue(class_fltest2)
        obj_ = class_fltest2()
        fake_cursor = obj_.cursor
        fake_cursor.setValueBuffer("date_field", "2001-02-01")
        self.assertTrue(isinstance(fake_cursor.valueBuffer("date_field"), qsa.Date))

    def test_pnsqlquery(self) -> None:
        """Test PNSqlQuery received values."""
        qry = qsa.FLSqlQuery()
        qry.setSelect("date_field")
        qry.setFrom("fltest")
        self.assertTrue(qry.exec_())
        self.assertTrue(qry.first())
        self.assertTrue(isinstance(qry.value("date_field"), qsa.Date))

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""
Simple parser for FacturaLUX Scripting Language (QSA).
"""
import optparse
import pprint
import sys
import math
import hashlib
import tempfile
import re
from typing import Any, Dict, List, Tuple, Optional, TextIO
from pineboolib.application.parsers import parser_qsa


import ply.yacc as yacc  # type: ignore
import ply.lex as lex  # type: ignore

from pineboolib.application.parsers.parser_qsa import token_rules, flex

TEMP_DIR = tempfile.gettempdir()

# Get the token map
tokens = token_rules.tokens
start = "source"  # pylint: disable=invalid-name

reserv = ["nonassoc"]
reserv += list(token_rules.reserved)

endoffile = None  # pylint: disable=invalid-name

hashes: List[Tuple[str, str]] = []
ranges: List[List[Any]] = []

SEEN_TOKENS = []
TOKE_LINES: Dict[int, int] = {}
LAST_LEXSPAN = None
INPUT_DATA: str = ""

precedence = (
    ("nonassoc", "EQUALS", "TIMESEQUAL", "DIVEQUAL", "MODEQUAL", "PLUSEQUAL", "MINUSEQUAL"),
    ("nonassoc", "MATHEXPRESSION"),
    ("nonassoc", "TERNARY"),
    ("left", "LOR", "LAND"),
    ("right", "LT", "LE", "GT", "GE", "EQ", "NE", "EQQ", "NEQ"),
    ("right", "LNOT"),
    ("left", "PLUS", "MINUS"),
    ("left", "TIMES", "DIVIDE", "MOD"),
    ("left", "OR", "AND", "XOR", "LSHIFT", "RSHIFT"),
)

LAST_OK_TOKEN: Any = None
ERROR_COUNT = 0
LAST_ERROR_TOKEN: Any = None
LAST_ERROR_LINE = -1
OK_COUNT = 0


def clean_no_python(data: str) -> str:
    """Remove NOPYTHON blocks."""
    return re.sub(r"\/\/___NOPYTHON\[\[.*?\/\/\]\]___NOPYTHON\s*", "", data, flags=re.DOTALL)


def clean_no_python_never(data: str) -> str:
    """Remove NOPYTHON_NEVER blocks."""
    return re.sub(
        r"\/\/___NOPYTHON_NEVER\[\[.*?\/\/\]\]___NOPYTHON_NEVER\s*", "", data, flags=re.DOTALL
    )


def clean_no_python_tags(data: str) -> str:
    """Remove no_python tags."""

    data = clean_no_python_never(data)
    if not parser_qsa.IGNORE_NO_PYTHON_TAGS:
        data = clean_no_python(data)

    return data


def cnvrt(val: str) -> str:
    """Convert special XML characters into XML entities."""
    val = str(val)
    val = val.replace("&", "&amp;")
    val = val.replace('"', "&quot;")
    val = val.replace("'", "&apos;")
    val = val.replace("<", "&lt;")
    val = val.replace(">", "&gt;")
    return val


def p_parse(token: Any) -> None:
    """Parse a single token."""
    global INPUT_DATA

    lexspan: List[int] = list(token.lexspan(0))
    # data = str(token.lexer.lexdata[lexspan[0]:lexspan[1]])
    # context = [
    #     str(token.lexer.lexdata[lexspan[0] - 32:lexspan[0]]),
    #     str(token.lexer.lexdata[lexspan[0]:lexspan[1]]),
    #     str(token.lexer.lexdata[lexspan[1]:lexspan[1] + 32]),
    # ]
    if len(lexspan) == 2:
        fromline = token.lineno(0)
        global endoffile  # pylint: disable=invalid-name
        endoffile = fromline, lexspan, token.slice[0]
    # print(repr(token.slice), context, lexspan)

    tok0: Dict[str, Any] = {
        "00-toktype": str(token.slice[0]),
        "02-size": lexspan,
        "50-contents": [{"01-type": s.type, "99-value": s.value} for s in token.slice[1:]],
    }
    token[0] = tok0

    for num in reversed(range(len(tok0["50-contents"]))):
        contents = tok0["50-contents"][num]
        if contents["01-type"] == tok0["00-toktype"]:
            # If the first element is same type, unpack left. This is for recursing lists
            # Makes tree calculation faster and less recursive.
            tok0["50-contents"][num : num + 1] = contents["99-value"]["50-contents"]

    numelems: int = len([s for s in token.slice[1:] if s.type != "empty" and s.value is not None])

    rspan = lexspan[0]
    if str(token.slice[0]) == "empty" or numelems == 0:
        token[0] = None
    else:
        rvalues = []
        for num, item in enumerate(token.slice[1:]):
            if item.type != "empty" and item.value is not None:
                val = None
                if isinstance(item.value, str):
                    val = token.lexspan(num + 1)[0] + len(item.value) - 1
                else:
                    val = token.lexspan(num + 1)[1]
                rvalues.append(val)
        rspan = max(rvalues)
    lexspan[1] = rspan

    # if str(token.slice[0]) == 'regexbody':
    #    token[0] = { "00-toktype": str(token.slice[0]) , "02-size" : lexspan,  "50-contents" :  INPUT_DATA[lexspan[0]:lexspan[1]+1] }

    # if str(token.slice[0]) == 'regex':
    #    print "\r\n",str(token.slice[0]) ,":" , INPUT_DATA[lexspan[0]:lexspan[1]+1]
    #    print "      " + "\n      ".join([ "%s(%r): %r" % (s.type, token.lexspan(n+1), s.value) for n,s in enumerate(token.slice[1:]) ])
    global SEEN_TOKENS, LAST_OK_TOKEN
    LAST_OK_TOKEN = token
    SEEN_TOKENS.append(
        (str(token.slice[0]), token.lineno(0), INPUT_DATA[lexspan[0] : lexspan[1] + 1])
    )
    global OK_COUNT
    OK_COUNT += 1
    if lexspan[0] not in TOKE_LINES:
        TOKE_LINES[lexspan[0]] = token.lexer.lineno
    global LAST_LEXSPAN
    LAST_LEXSPAN = lexspan


def p_error(text: Any) -> Any:
    """Process and report errors in parsing."""
    global ERROR_COUNT
    global OK_COUNT
    global LAST_ERROR_TOKEN
    global LAST_ERROR_LINE, SEEN_TOKENS, LAST_OK_TOKEN
    debug = False  # Poner a True para toneladas de debug.
    # if ERROR_COUNT == 0: print
    if text is not None:
        if LAST_ERROR_TOKEN is None or text.lexpos != getattr(LAST_ERROR_TOKEN, "lexpos", None):
            if abs(LAST_ERROR_LINE - text.lineno) > 4 and OK_COUNT > 1 and ERROR_COUNT < 4:
                ERROR_COUNT += 1
                try:
                    print_context(text)
                except Exception:
                    pass
                if debug:
                    ERROR_COUNT += 20  # no imprimir mas de un error en debug.
                    print()
                    for tokname, tokln, tokdata in SEEN_TOKENS[-32:]:
                        if tokln == text.lineno:
                            print(tokname, tokdata)
                    if LAST_OK_TOKEN:
                        print(repr(LAST_OK_TOKEN[0]))
                        for sli_ in LAST_OK_TOKEN.slice[:]:
                            print(
                                ">>>", sli_.lineno, repr(sli_), pprint.pformat(sli_.value, depth=3)
                            )
                LAST_ERROR_LINE = text.lineno
            elif abs(LAST_ERROR_LINE - text.lineno) > 1 and OK_COUNT > 1:
                LAST_ERROR_LINE = text.lineno
            parser.errok()
            OK_COUNT = 0
            return

    OK_COUNT = 0
    if text is None:
        if LAST_ERROR_TOKEN != "EOF":
            print("ERROR: End of the file reached.")
            global endoffile  # pylint: disable=invalid-name
            print("Last data:", endoffile)

            if LAST_LEXSPAN:
                try:
                    print("HINT: Last lexspan:", LAST_LEXSPAN)
                    print("HINT: Last line:", TOKE_LINES[LAST_LEXSPAN[0]])
                except Exception as error:
                    print("ERROR:", error)
        LAST_ERROR_TOKEN = "EOF"
        return text
    text = parser.token()
    parser.restart()
    LAST_ERROR_TOKEN = text
    return text


p_parse.__doc__ = """

    inlinestoreinstruction  : PLUSPLUS variable
                            | MINUSMINUS variable
                            | variable PLUSPLUS
                            | variable MINUSMINUS

    exprval : constant
            | variable
            | funccall
            | error

    identifier : ID

    dictobject_value : LBRACE RBRACE
                     | LBRACE dictobject_value_elemlist RBRACE
                     | LBRACE dictobject_value_elemlist COMMA RBRACE

    dictobject_value_elemlist : dictobject_value_elem
                              | dictobject_value_elemlist COMMA dictobject_value_elem

    dictobject_value_elem : exprval COLON expression

    base_expression     : ternary_operator
                        | inlinestoreinstruction
                        | parentheses
                        | unary_operator
                        | new_operator
                        | typeof_operator
                        | dictobject_value
                        | exprval
                        | base_expression mathoperator base_expression      %prec MATHEXPRESSION
                        | base_expression cmp_symbol base_expression        %prec MATHEXPRESSION
                        | base_expression boolcmp_symbol base_expression    %prec MATHEXPRESSION





    parentheses         : LPAREN base_expression RPAREN
                        | LPAREN variable_1 RPAREN

    unary_operator      : LNOT base_expression
                        | MINUS base_expression
                        | PLUS base_expression

    new_operator        : NEW funccall_1
                        | NEW identifier

    typeof_operator     : TYPEOF parentheses

    expression  : base_expression
                | funcdeclaration_anon
                | funcdeclaration_anon_exec
                | LPAREN expression RPAREN
                | error

    case_cblock_list  :  case_block
    case_cblock_list  :  case_cblock_list case_block

    case_block  :  CASE expression COLON statement_list

    case_default    :  DEFAULT COLON statement_list

    case_block_list  :  empty
    case_block_list  :  case_default
    case_block_list  :  case_cblock_list
    case_block_list  :  case_cblock_list case_default

    source_element  : docstring
                    | vardeclaration
                    | classdeclaration
                    | funcdeclaration
                    | funcdeclaration_anon
                    | funcdeclaration_anon_exec

    source  : source_element
    source  : source source_element
            | statement_list


    basicsource     : statement_list
                    | empty

    statement   : instruction
                | vardeclaration
                | ifstatement
                | whilestatement
                | dowhilestatement
                | withstatement
                | forstatement
                | forinstatement
                | switch
                | trycatch

    statement_list      : statement_list statement

    statement_list      : statement

    statement_list      : LBRACE statement_list RBRACE

    statement_list      : LBRACE RBRACE
    statement_list      : empty

    optvartype  : COLON ID
                | empty

    vardeclaration  :  VAR vardecl_list SEMI
                    |  CONST vardecl_list SEMI
    vardeclaration  :  VAR vardecl_list
                    |  CONST vardecl_list
                    |  STATIC VAR vardecl_list

    vardecl  :  ID optvartype EQUALS expression
             |  vardecl EQUALS expression
    vardecl  :  ID optvartype

    vardecl_list    : vardecl
                    | vardecl_list COMMA vardecl

    arglist : vardecl_list
            |

    funcdeclaration : FUNCTION ID LPAREN arglist RPAREN optvartype LBRACE basicsource RBRACE
    funcdeclaration : STATIC FUNCTION ID LPAREN arglist RPAREN optvartype LBRACE basicsource RBRACE
    funcdeclaration_anon : FUNCTION LPAREN arglist RPAREN LBRACE basicsource RBRACE
                         | FUNCTION LPAREN RPAREN LBRACE basicsource RBRACE
    funcdeclaration_anon_exec : funcdeclaration_anon LPAREN RPAREN
                              | funcdeclaration_anon LPAREN arglist RPAREN

    callarg     : expression

    callargs    : callarg
                | callargs COMMA callarg

    varmemcall  : variable_1
                | funccall_1
                | member_call
                | member_var
                | base_expression

    member_var  : varmemcall PERIOD variable_1

    member_call : LPAREN member_var RPAREN PERIOD funccall_1

    member_call : varmemcall PERIOD funccall_1
    member_call : LPAREN member_call RPAREN PERIOD funccall_1

    funccall    : funccall_1
                | member_call
                | LPAREN member_call RPAREN
                | LPAREN funccall_1 RPAREN
                | LPAREN error RPAREN

    funccall_1  : ID LPAREN callargs RPAREN
                | ID LPAREN RPAREN
                | TYPEOF LPAREN callargs RPAREN
                | array_member LPAREN callargs RPAREN
                | array_member LPAREN RPAREN

    mathoperator    : PLUS
                    | MINUS
                    | TIMES
                    | DIVIDE
                    | MOD
                    | XOR
                    | OR
                    | LSHIFT
                    | RSHIFT
                    | AND

    variable    : variable_1
                | member_var
                | LPAREN variable_1 RPAREN
                | LPAREN member_var RPAREN

    variable_1  : identifier
                | array_member

    array_member : variable_1 LBRACKET expression RBRACKET
                 | funccall_1 LBRACKET expression RBRACKET

        updateoperator : EQUALS
                       | PLUSEQUAL
                       | MINUSEQUAL
                       | MODEQUAL
                       | DIVEQUAL
                       | TIMESEQUAL

        updateinstruction : variable updateoperator expression
                          | variable updateoperator updateinstruction

        deleteinstruction   : DELETE variable

        storeinstruction    : inlinestoreinstruction
                            | updateinstruction
                            | deleteinstruction


    flowinstruction : RETURN expression
                    | RETURN expression COMMA
                    | THROW expression
                    | RETURN
                    | BREAK
                    | CONTINUE

    instruction : base_instruction SEMI
                | SEMI
                | base_instruction
                | funcdeclaration
                | error SEMI

    callinstruction : funccall
                    | variable

    base_instruction : storeinstruction
                | callinstruction
                | flowinstruction

    varorcall : variable
              | funccall
              | base_expression

    optextends  : EXTENDS ID
                | empty

    classdeclaration   : CLASS ID optextends LBRACE classdeclarationsource RBRACE

    classdeclarationsource  : vardeclaration
                            | funcdeclaration
                            | classdeclarationsource vardeclaration
                            | classdeclarationsource funcdeclaration
                            | SEMI
                            | classdeclarationsource SEMI

    docstring   : DOCSTRINGOPEN AT ID COMMENTCLOSE
                | DOCSTRINGOPEN AT ID ID COMMENTCLOSE

    list_constant   : LBRACKET RBRACKET
    list_constant   : LBRACKET callargs RBRACKET
    list_constant   : LBRACKET callargs COMMA RBRACKET

    constant : ICONST
                | FCONST
                | CCONST
                | SCONST
                | regex
                | list_constant

    regex : DIVIDE regexbody DIVIDE regexflags
          | DIVIDE regexbody COMMENTCLOSE regexflags

    regexbody   : regexchar
                | regexbody regexchar

    regexchar :  LPAREN
              | RPAREN
              | ID
              | COMMA
              | XOR
              | LBRACKET
              | RBRACKET
              | ICONST
              | PLUS
              | MINUS
              | LBRACE
              | RBRACE
              | DOLLAR
              | SQOUTE
              | DQOUTE
              | PERIOD
              | BACKSLASH
              | CONDITIONAL1
              | EQUALS
              | OR
              | SCONST
              | SEMI
              | MOD
              | AT
              | COLON
              | error

    regexflags : ID
               | empty


    statement_block : statement
                    | LBRACE statement_list RBRACE
                    | LBRACE RBRACE

    optelse : ELSE statement_block
            | empty

    cmp_symbol  : LT
                | LE
                | GT
                | GE
                | EQ
                | NE
                | EQQ
                | NEQ
                | IN

    boolcmp_symbol  : LOR
                    | LAND

    condition   : expression
                | error

    ifstatement : IF LPAREN condition RPAREN statement_block optelse

    whilestatement  : WHILE LPAREN condition RPAREN statement_block
    dowhilestatement  : DO statement_block WHILE LPAREN condition RPAREN SEMI
                      | DO statement_block WHILE LPAREN condition RPAREN

    withstatement   : WITH LPAREN variable RPAREN statement_block
                    | error

    storeormember   : storeinstruction
                    | member_var

    for_initialize  : storeinstruction
                    | VAR vardecl
                    | vardecl
                    | for_initialize COMMA for_initialize
                    | empty

    for_compare     : expression
                    | empty

    for_increment   : storeormember
                    | for_increment COMMA for_increment
                    | empty


    forstatement    : FOR LPAREN for_initialize SEMI for_compare SEMI for_increment RPAREN statement_block
                    | error

    forinstatement  : FOR LPAREN for_initialize IN varorcall RPAREN statement_block
                    | FOR LPAREN variable IN varorcall RPAREN statement_block
                    | error

    switch  : SWITCH LPAREN condition RPAREN LBRACE case_block_list RBRACE

    optid   : ID
            | empty

    trycatch    : TRY statement_block CATCH LPAREN optid RPAREN statement_block

    ternary_operator    : base_expression CONDITIONAL1 base_expression COLON base_expression %prec TERNARY

    empty :
    """


# Build the grammar


parser = yacc.yacc(
    method="LALR",
    debug=0,
    optimize=1,
    write_tables=1,
    debugfile="%s/yaccdebug.txt" % TEMP_DIR,
    outputdir="%s/" % TEMP_DIR,
)

# parser = yacc.yacc(method='LALR', debug=1,
#                   optimize=0, write_tables=0, debugfile='%s/yaccdebug.txt' % TEMP_DIR, outputdir='%s/' % TEMP_DIR)

# profile.run("yacc.yacc(method='LALR')")


def print_context(token: Any) -> None:
    """Report errors in console when parsing fails."""
    global INPUT_DATA
    if token is None:
        return
    last_cr = INPUT_DATA.rfind("\n", 0, token.lexpos)
    next_cr = INPUT_DATA.find("\n", token.lexpos)
    column = token.lexpos - last_cr
    column1 = token.lexpos - last_cr
    while column1 < 16:
        column1 = token.lexpos - last_cr
        last_cr = INPUT_DATA.rfind("\n", 0, last_cr - 1)

    print(INPUT_DATA[last_cr:next_cr].replace("\t", " "))
    print((" " * (column - 1)) + "^", column, "#ERROR#", token)


def my_tokenfunc(*args, **kwargs) -> Any:
    """Tokenizer function."""
    # print("Call token:" ,args, kwargs)
    ret = lex.lexer.token(*args, **kwargs)
    # print "Return (",args, kwargs,") = " , ret
    return ret


def print_tokentree(token: Any, depth: int = 0) -> None:
    """Debug output. Unused."""
    print("  " * depth, token.__class__, "=", token)

    if str(token.__class__) == "ply.yacc.YaccProduction":
        print(token.lexer)
        for token in token.slice:
            if token.value == token:
                continue
            print("  " * (depth + 1), token.type, end=" ")
            try:
                print(token.lexpos, end=" ")
                print(token.endlexpos, end=" ")
            except Exception:
                pass
            print()

            print_tokentree(token.value, depth + 1)


def calctree(
    obj: Dict[str, Any],
    depth: int = 0,
    num: List[str] = [],
    otype: str = "source",
    alias_mode: int = 1,
) -> Dict[str, Any]:
    """Extract parsed AST and generate a custom structure for later XML generation."""
    # if depth > 5: return
    # source_data = [
    #     'source',
    #     'source_element',
    #     'statement_list',
    #     'statement',
    #     'classdeclarationsource',
    #     'vardecl_list',
    # ]
    final_obj = {}
    final_obj["range"] = obj["02-size"]
    has_data = 0
    has_objects = 0
    contentlist = []
    ctype_alias: Dict[str, str]
    if alias_mode == 0:
        ctype_alias = {}
    elif alias_mode == 1:
        ctype_alias = {
            "member_var": "member",
            "member_call": "member",
            "variable_1": "variable",
            "funccall_1": "funccall",
            "flowinstruction": "instruction",
            "storeequalinstruction": "instruction",
            "vardecl": "vardeclaration",
            # "vardecl_list" : "vardeclaration",
        }
    else:
        raise ValueError("alias_mode unknown")

    if otype in ctype_alias:
        otype = ctype_alias[otype]
    # print " " * depth , obj['02-size']
    for num_, content in enumerate(obj["50-contents"]):
        if not isinstance(content, dict):
            print("ERROR: content is not a dict!:", repr(content))
            print(".. obj:", repr(obj))
            raise TypeError("content is not a dict")

        ctype = content["01-type"]
        value = content["99-value"]
        if ctype in ctype_alias:
            ctype = ctype_alias[ctype]
        # if ctype in source_data:
        #    if depth == 0: print "--"
        #    print_tree(value,depth,num)
        #    continue
        # print " " * depth , "%s:" % ".".join(num+[str(num_)]), ctype,

        if isinstance(value, dict):
            # print "*"
            # FIXME: Esto o no parsea todos los elementos o hace stackoverflow. problematico para programas largos
            if depth < 300:
                try:
                    tree_obj = calctree(
                        value, depth + 1, num + [str(num_)], ctype, alias_mode=alias_mode
                    )
                except Exception as error:
                    print("ERROR: trying to calculate member:", error)
                    continue
            else:
                print("PANIC: *** Stack overflow trying to calculate member %d on:" % num_, ctype)
                continue

            if type(tree_obj) is dict:
                if (tree_obj["has_data"] or alias_mode == 0) and ctype != otype:
                    contentlist.append([ctype, tree_obj])
                    has_objects += 1
                else:
                    contentlist += tree_obj["content"]
                    has_data += tree_obj["has_data"]
                    has_objects += tree_obj["has_objects"]
        else:
            # print "=", repr(value)
            contentlist.append([ctype, value])
            has_data += 1

    final_obj["content"] = contentlist
    final_obj["has_data"] = has_data
    final_obj["has_objects"] = has_objects

    return final_obj


def printtree(
    tree: Dict[str, Any],
    depth: int = 0,
    otype: str = "source",
    mode: Optional[str] = None,
    output: TextIO = sys.stdout,
) -> Tuple[Any, Any, Any]:
    """Export AST into different formats, mainly used for XML export."""
    global hashes, ranges  # pylint: disable=invalid-name
    if depth == 0:
        hashes = []
        ranges = []

    sep = "    "
    marginblocks = {
        "classdeclaration": 1,
        "funcdeclaration": 1,
        "statement_block": 1,
        # "instruction" : 1,
    }
    closingtokens = ["RBRACE", "RPAREN", "RBRACKET", "SEMI"]
    nuevalinea = False
    name = ""
    lines: List[str] = []
    L = 0  # pylint: disable=invalid-name

    for ctype, value in tree["content"]:
        if nuevalinea:
            if ctype in closingtokens:
                nuevalinea = False

            else:
                for item in range(int(math.ceil(L / 2.0))):
                    lines.append(sep * depth)
                nuevalinea = False

        if isinstance(value, dict):
            if ctype == otype:
                tname, tlines, trange = printtree(value, depth, ctype)
                if name == "" and tname:
                    name = tname

                lines += tlines
            else:
                L = 0  # pylint: disable=invalid-name
                if ctype in marginblocks:
                    L = marginblocks[ctype]  # pylint: disable=invalid-name

                for i in range(int(math.floor(L / 2.0))):
                    lines.append(sep * depth)
                tname, tlines, trange = printtree(value, depth + 1, ctype)
                # lines.append(sep * depth + "<!-- %d -->" % (len("".join(tlines))))

                if value["has_data"] > 0 and value["has_objects"] == 0 and False:
                    pass
                    # Do it inline!
                    # if value["has_data"] == 1 and tname:
                    #     lines.append(sep * depth + '<%s id="%s" />' % (ctype, tname))
                    # else:
                    #     txt = "".join([x.strip() for x in tlines])
                    #     lines.append(sep * depth + "<%s>%s</%s>" % (ctype, txt, ctype))
                else:
                    attrs = []
                    if tname:
                        attrs.append(("id", tname))

                    txtinline = "".join([line.strip() for line in tlines])

                    # if len(tlines)>1:
                    txthash = hashlib.sha1(txtinline.encode("utf8")).hexdigest()[:16]
                    # hashes.append(("depth:",depth,"hash:",txthash,"element:",ctype+":"+tname))
                    hashes.append((txthash, ctype + ":" + tname + "(%d)" % len(txtinline)))
                    ranges.append([depth, txthash] + trange + [ctype + ":" + tname, len(txtinline)])
                    # ,"start:",trange[0],"end:",trange[1]))
                    # attrs.append(("start",trange[0]))
                    # attrs.append(("end",trange[1]))
                    # attrs.append(("hash",txthash))

                    txtattrs = ""
                    for name1, val1 in attrs:
                        txtattrs += ' %s="%s"' % (name1, cnvrt(val1))

                    lines.append(sep * depth + "<%s%s>" % (ctype, txtattrs))
                    if depth > 50:
                        lines.append(sep * (depth + 1) + "...")
                    else:
                        if len(txtinline) < 80:
                            lines.append(sep * (depth + 1) + txtinline)
                        else:
                            lines += tlines
                    if txtattrs:
                        txtattrs = "<!--%s -->" % txtattrs
                    lines.append(sep * depth + "</%s>" % (ctype))

                    nuevalinea = True
        else:
            if ctype == "ID" and name == "":
                name = value
            if ctype in token_rules.token_literals:
                lines.append(sep * depth + '<%s value="%s" />' % (ctype, cnvrt(value)))
            else:
                lines.append(sep * depth + "<%s />" % (ctype))

    if mode == "hash":
        # print "\n".join(lines)
        for row in sorted(ranges):
            output.write("\t".join([str(x) for x in row]))
            output.write("\n")
            output.flush()
    elif mode == "xml":
        for strrow in lines:
            output.write(strrow)
            output.write("\n")
            output.flush()

    return name, lines, tree["range"]


def parse(data: str, clean: bool = True) -> Optional[Dict[str, Any]]:
    """Parse QS String."""

    global INPUT_DATA
    global ERROR_COUNT
    global SEEN_TOKENS

    if clean:
        data = clean_no_python_tags(data)

    SEEN_TOKENS[:] = []
    parser.error = 0
    INPUT_DATA = data
    flex.lexer.lineno = 1
    ERROR_COUNT = 0
    parser_ = parser.parse(data, debug=0, tracking=1, tokenfunc=my_tokenfunc)
    if ERROR_COUNT > 0:
        print("ERRORS (%d)" % ERROR_COUNT)
    if parser_ is not None:
        try:
            parser_["error_count"] = ERROR_COUNT
        except Exception as error:
            print(error)
            return None

        if parser.error:
            return None

    return parser_


def main() -> None:
    """Manage direct script calls for flscriptparse. Deprecated."""
    global start  # pylint: disable=invalid-name
    parser = optparse.OptionParser()
    parser.add_option(
        "-O",
        "--output",
        dest="output",
        default="none",
        help="Set output TYPE: xml|hash",
        metavar="TYPE",
    )
    parser.add_option("--start", dest="start", default=None, help="Set start block", metavar="STMT")
    parser.add_option(
        "-q",
        "--quiet",
        action="store_false",
        dest="verbose",
        default=True,
        help="don't print status messages to stdout",
    )
    parser.add_option(
        "--optdebug",
        action="store_true",
        dest="optdebug",
        default=False,
        help="debug optparse module",
    )
    parser.add_option(
        "--debug",
        action="store_true",
        dest="debug",
        default=False,
        help="prints lots of useless messages",
    )

    (options, args) = parser.parse_args()
    if options.optdebug:
        print(options, args)
    if options.start:
        start = options.start
        print("Start setted to:", start)

    def do_it():
        if options.output == "none":
            return
        tree_data = calctree(prog)
        if options.output == "hash":
            printtree(tree_data, mode="hash")
        elif options.output == "xml":
            printtree(tree_data, mode="xml")
        elif options.output == "file":
            f1_hash = open(filename + ".hash", "w")
            printtree(tree_data, mode="hash", output=f1_hash)
            f1_hash.close()

            f1_xml = open(filename + ".xml", "w")
            printtree(tree_data, mode="xml", output=f1_xml)
            f1_xml.close()
        elif options.output == "yaml":
            import yaml  # type: ignore [import]

            print(yaml.safe_dump(tree_data["content"]))

        else:
            print("Unknown outputmode", options.output)

    prog: Dict[str, Any]
    if len(args) > 0:
        for filename in args:
            fs = filename.split("/")  # pylint: disable=invalid-name
            sys.stderr.write("Loading %s ..." % fs[-1])
            sys.stderr.flush()
            data = open(filename).read()
            sys.stderr.write(" parsing ...")
            sys.stderr.flush()
            prog_ = parse(data)
            if prog_ is not None:
                prog = prog_
                sys.stderr.write(" formatting ...")
                sys.stderr.flush()
                do_it()
                sys.stderr.write(" Done.\n")
                sys.stderr.flush()

    else:
        line = ""
        while 1:
            try:
                line1 = input("flscript> ")
                if line1.startswith("#"):
                    comm = line1[1:].split(" ")
                    if comm[0] == "setstart":
                        start = comm[1]
                        print("Start setted to:", start)
                    if comm[0] == "parse":
                        print()
                        line = ""
                        prog_ = parse(line)
                        if prog_ is None:
                            continue
                        prog = prog_

                else:
                    line += line1
            except EOFError:
                break
            line += "\n"
        print()
        prog_ = parse(line)
        if prog_ is not None:
            prog = prog_
            do_it()
    """
    import yaml

    print yaml.dump(tree_data)
    """
    # print_tokentree(prog)

    # for varName in prog.byDefName:
    #    var = prog.byDefName[varName]
    #    print "%-15s / %-15s > " % var.type  , varName

    # import tests.ifaceclass
    # tests.ifaceclass.do_test(prog)


if __name__ == "__main__":
    main()
